from joblib import cpu_count
import traceback
import dill as pickle
import os
from datetime import datetime
# This file defines variables that can be accessed globally.

root = ""
app_name = "scNodes"
version = "1.0.41"
license = "GNU GPL v3"
logpath = "log.txt"
nodes = list()
active_node = None
focused_node = None
any_change = False  # top-level change flag
next_active_node = None
node_move_requested = [0, 0]
camera_move_requested = [0, 0]

window_width = 1100
window_height = 700
iv_window_width = 600
iv_window_height = 700

profiling = False

error_msg = None
error_new = True
error_obj = None
error_logged = False

active_connector = None
active_connector_parent_node = None
active_connector_hover_pos = [0, 0]
connector_released = False
connector_delete_requested = False

n_cpus_max = cpu_count()
n_cpus = n_cpus_max
batch_size = n_cpus_max

image_viewer = None
node_editor = None
correlation_editor = None
node_editor_relink = False
correlation_editor_relink = False
pickle_temp = dict()


## 221221 correlation editor vars & related
active_editor = 0  # 0 for node editor, 1 for correlation
ce_frames = list()
ce_active_frame = None
ce_clear_colour = (1.0, 1.0, 1.0, 1.0)
ce_default_pixel_size = 64.0
ce_flip_on_load = False
ce_flip_mrc_on_load = False
ce_selected_position = [0, 0]

def set_active_node(node, keep_active=False):
    global focused_node, active_node, next_active_node
    """
    :param node: Node type object or None.
    :param keep_active: bool. If True, the node is not just set as active_node, but also as the focused_node - this means that the node is kept in focus until it is manually unfocused by the user (in node right click context menu).
    :return: False if no change; i.e. when input node is already the active node. True otherwise.
    """
    if keep_active:
        focused_node = node
        next_active_node = node
        return True
    elif focused_node is None:
        if active_node is not None:
            if active_node == node:
                return False
        if node is None:
            active_node = None
            return True
        else:
            next_active_node = node
            return True
    else:
        return False


def set_error(error_object, error_message):
    global error_msg, error_obj, error_new, error_logged
    error_msg = error_message + "\n\n"
    error_msg += "".join(traceback.TracebackException.from_exception(error_object).format())
    print(error_msg)
    error_obj = error_object
    error_new = True
    error_logged = False

def save_project(filename):
    if filename[-5:] == '.srnp':
        filename = filename[:-5]
    with open(filename+'.srnp', 'wb') as pickle_file:
        pickle.dump([nodes, ce_frames], pickle_file)

def load_project(filename):
    global nodes, ce_frames, correlation_editor_relink, node_editor_relink
    try:
        with open(filename, 'rb') as pickle_file:
            data = pickle.load(pickle_file)
            nodes = data[0]
            ce_frames = data[1]
            node_editor_relink = True
            correlation_editor_relink = True
    except Exception as e:
        set_error(e, "Error loading project")

def write_to_log(text):
    with open(root+logpath, "a") as f:
        f.write("\n\n ____________________ \n\n")
        f.write(text)

def start_log():
    if os.path.exists(root+logpath):
        os.remove(root+logpath)
    with open(root+logpath, "a") as f:
        f.write(app_name+" version "+version+" "+license+"\n"+datetime.now().strftime("%m/%d/%Y, %H:%M:%S"))

COLOUR_WINDOW_BACKGROUND = (0.94, 0.94, 0.94, 0.94)
COLOUR_PANEL_BACKGROUND = (0.94, 0.94, 0.94, 0.94)
COLOUR_TITLE_BACKGROUND = (0.87, 0.87, 0.83, 0.96)
COLOUR_TITLE_BACKGROUND_LIGHT = (0.96, 0.96, 0.93, 0.93)
COLOUR_FRAME_BACKGROUND = (0.87, 0.87, 0.83, 0.96)
COLOUR_FRAME_ACTIVE = (0.91, 0.91, 0.86, 0.94)
COLOUR_FRAME_DARK = (0.83, 0.83, 0.76, 0.94)
COLOUR_FRAME_EXTRA_DARK = (0.76, 0.76, 0.71, 0.94)
COLOUR_MAIN_MENU_BAR = (0.882, 0.882, 0.882, 0.94)
COLOUR_MAIN_MENU_BAR_TEXT = (0.0, 0.0, 0.0, 0.94)
COLOUR_MAIN_MENU_BAR_HILIGHT = (0.96, 0.95, 0.92, 0.94)
COLOUR_MENU_WINDOW_BACKGROUND = (0.96, 0.96, 0.96, 0.94)
COLOUR_DROP_TARGET = COLOUR_FRAME_DARK
COLOUR_HEADER = COLOUR_FRAME_DARK
COLOUR_HEADER_ACTIVE = COLOUR_FRAME_ACTIVE
COLOUR_HEADER_HOVERED = COLOUR_FRAME_EXTRA_DARK
COLOUR_TEXT = (0.0, 0.0, 0.0, 1.0)
COLOUR_TEXT_FADE = COLOUR_FRAME_EXTRA_DARK
WINDOW_ROUNDING = 5.0
CONTEXT_MENU_SIZE = (200, 98)
ERROR_WINDOW_HEIGHT = 80
COLOUR_ERROR_WINDOW_BACKGROUND = (0.84, 0.84, 0.84, 1.0)
COLOUR_ERROR_WINDOW_HEADER = (0.7, 0.7, 0.7, 1.0)
COLOUR_ERROR_WINDOW_HEADER_NEW = (0.35, 0.35, 0.35, 1.0)
COLOUR_ERROR_WINDOW_TEXT = (0.0, 0.0, 0.0, 1.0)
COLOUR_CM_WINDOW_TEXT = (0.0, 0.0, 0.0, 1.0)
COLOUR_CM_OPTION_HOVERED = (1.0, 1.0, 1.0, 1.0)

TOOLTIP_APPEAR_DELAY = 1.0
TOOLTIP_HOVERED_TIMER = 0.0
TOOLTIP_HOVERED_START_TIME = 0.0

CE_WIDGET_ROUNDING = 50.0

start_log()


from scNodes.core.node import *
from pystackreg import StackReg
import cv2


def create():
    return RegisterNode()


class RegisterNode(Node):
    title = "Registration"
    group = "Image processing"  # default groups: "Data IO", "Image processing", "Reconstruction", "Custom"
    colour = (68 / 255, 177 / 255, 209 / 255, 1.0)
    sortid = 100
    METHODS = ["TurboReg", "ORB"]
    REFERENCES = ["Input image", "Template frame", "Consecutive pairing"]
    INTERPOLATION_OPTIONS = ["Nearest neighbour", "Bilinear", "Biquadratic", "Bicubic", "Biquartic", "Biquintic"]
    EDGE_FILL_OPTIONS = ["Zero", "Repeat", "Reflect"]
    edge_fill_options_scipy_argument = ['constant', 'edge', 'reflect']

    def __init__(self):
        super().__init__()  # Was: super(LoadDataNode, self).__init__()
        self.size = 230

        # Set up connectable attributes
        self.dataset_in = ConnectableAttribute(ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.INPUT, parent = self)
        self.dataset_out = ConnectableAttribute(ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.OUTPUT, parent = self)
        self.image_in = ConnectableAttribute(ConnectableAttribute.TYPE_IMAGE, ConnectableAttribute.INPUT, parent = self)

        # Set up node-specific vars
        self.buffer_last_output = True
        self.register_method = 0
        self.reference_method = 1
        self.reference_image = None
        self.frame = 0
        self.roi = [0, 0, 0, 0]

        # StackReg vars
        self.sr = StackReg(StackReg.TRANSLATION)

        # CV vars
        self.orb = None
        self.orb_n_actual = 0
        self.orb_n_requested = 500
        self.orb_keep = 0.7
        self.orb_confidence = 0.99
        self.orb_method = 0

        # Advanced
        self.interpolation = 1
        self.edge_fill = 0
        self.preserve_range = False

    def render(self):
        ## TODO: refresh input image when input image was updated.
        if super().render_start():
            self.dataset_out.render_start()
            self.dataset_in.render_start()
            self.dataset_out.render_end()
            self.dataset_in.render_end()

            imgui.spacing()
            imgui.separator()
            imgui.spacing()

            _c, self.use_roi = imgui.checkbox("use ROI", self.use_roi)
            self.any_change = self.any_change or _c
            self.tooltip("When 'use ROI' is active, the translation required to register the image is determined based\n"
                         "on the data in the ROI only. This can help speed up registration, as well as avoid errors\n"
                         "such as occur when the registrator non-constant image features like blinking particles are\n"
                         "present in the full image.")
            imgui.push_item_width(140)
            _method_changed, self.register_method = imgui.combo("Method", self.register_method, RegisterNode.METHODS)
            if self.register_method == 0:
                imgui.same_line()
                imgui.button("?", 19, 19)
                self.tooltip("This method uses pyStackReg, a TurboReg wrapper by Gregor Lichtner (@glichtner). Original publication:\nP. Thévenaz, U. E. Ruttimann, M. Unser (1998) A Pyramid Approach to Subpixel Registration Based on Intensity. \nIEEE Trans. Image Process. 7:1:27-41 doi: 10.1109/83.650848")
            _reference_changed, self.reference_method = imgui.combo("Reference", self.reference_method, RegisterNode.REFERENCES)
            imgui.pop_item_width()

            _frame_changed = False
            if self.reference_method == 1:
                imgui.push_item_width(50)
                _frame_changed, self.frame = imgui.input_int("Template frame", self.frame, 0, 0)
                imgui.pop_item_width()
            if self.reference_method == 0:
                imgui.new_line()
                self.image_in.render_start()
                self.image_in.render_end()

            if self.register_method == 0:
                pass  # no options (yet) for TurboReg
            elif self.register_method == 1:
                imgui.push_item_width(110)
                _c, self.orb_method = imgui.combo("Estimator", self.orb_method, ["Random sample consensus", "Least median of squares"])
                self.any_change = self.any_change or _c
                imgui.pop_item_width()
                imgui.push_item_width(90)
                _c, self.orb_n_requested = imgui.input_int("# features", self.orb_n_requested, 100, 100)
                self.any_change = self.any_change or _c
                _c, self.orb_keep = imgui.input_float("% matches to keep", self.orb_keep, format = "%.2f")
                self.any_change = self.any_change or _c
                _c, self.orb_confidence = imgui.input_float("confidence", self.orb_confidence, format = "%.2f")
                self.any_change = self.any_change or _c
                if _c:
                    self.orb_confidence = min([1.0, max([0.1, self.orb_confidence])])
                imgui.pop_item_width()

            self.any_change = self.any_change or _method_changed or _reference_changed or _frame_changed
            if self.any_change:
                self.reference_image = None

            header_expanded, _ = imgui.collapsing_header("Advanced", None)
            if header_expanded:
                self.render_advanced()

            super().render_end()

    def render_advanced(self):
        imgui.push_item_width(110)
        _c, self.interpolation = imgui.combo("Interpolation", self.interpolation, RegisterNode.INTERPOLATION_OPTIONS)
        self.any_change = self.any_change or _c
        _c, self.edge_fill = imgui.combo("Edges", self.edge_fill, RegisterNode.EDGE_FILL_OPTIONS)
        self.any_change = self.any_change or _c
        imgui.pop_item_width()
        Node.tooltip("Set how to fill pixels falling outside of the original image. Edge: clamp the edge\n"
                     "values, Reflect: reflect the image along the boundaries of the original image.")
        _c, self.preserve_range = imgui.checkbox("Preserve range", self.preserve_range)
        self.any_change = self.any_change or _c
        Node.tooltip("When checked, the intensity range of the output, registered image is fixed as the\n"
                     "same range as that of the original image.")

    def get_image_impl(self, idx=None):
        data_source = self.dataset_in.get_incoming_node()
        if data_source:
            input_img = data_source.get_image(idx)
            if self.reference_image is None:
                # Get reference frame according to specified pairing method
                if self.reference_method == 2:
                    self.reference_image = data_source.get_image(idx - 1).load()
                elif self.reference_method == 0:
                    self.reference_image = self.image_in.get_incoming_node().get_image(idx=None).load()
                elif self.reference_method == 1:
                    self.reference_image = data_source.get_image(self.frame).load()

            # Perform registration according to specified registration method

            if self.reference_image is not None:
                template = self.reference_image
                image = input_img.load()
                if self.use_roi:
                    template = template[self.roi[1]:self.roi[3], self.roi[0]:self.roi[2]]
                    image = image[self.roi[1]:self.roi[3], self.roi[0]:self.roi[2]]
                if self.register_method == 0:
                    tmat = self.sr.register(template, image)
                    input_img.translation = [tmat[0][2], tmat[1][2]]
                elif self.register_method == 1:
                    ## make new ORB if necessary
                    if self.orb_n_actual != self.orb_n_requested:
                        self.orb = cv2.ORB_create(nfeatures=self.orb_n_requested)
                    tmat = self._orb_register(template, image, keep=self.orb_keep, confidence=self.orb_confidence,
                                              method=self.orb_method)
                    input_img.translation = tmat
            else:
                cfg.set_error(Exception(), "RegisterNode: reference image was None.")

            if self.reference_method == 2:
                self.reference_image = None
            input_img.bake_transform(interpolation=self.interpolation, edges=RegisterNode.edge_fill_options_scipy_argument[self.edge_fill], preserve_range=self.preserve_range)
            return input_img

    def pre_save_impl(self):
        self.reference_image = None

    def _orb_register(self, ref, img, keep=0.7, confidence=0.99, method=0):
        """

        :param ref: reference image
        :param img: image to be registered
        :param keep: the fraction (0.0 to 1.0) of matches to use to estimate the transformation - e.g. 0.1: keep top 10% matches
        :param confidence: confidence parameter of cv2.estimateAffinePartial2d
        :param method: 0 for RANSAC, 1 for LMEDS
        :return: list with two elements: x shift and y shift, in pixels.
        """
        cmin = np.amin(img)
        cmax = np.amax(img)
        ref8 = 255 * (ref - cmin) / (cmax - cmin)
        ref8 = ref8.astype(np.uint8)
        img8 = 255 * (img - cmin) / (cmax - cmin)
        img8 = img8.astype(np.uint8)

        kp1, d1 = self.orb.detectAndCompute(ref8, None)
        kp2, d2 = self.orb.detectAndCompute(img8, None)

        matcher = cv2.BFMatcher(cv2.NORM_HAMMING, crossCheck=True)
        matches = list(matcher.match(d1, d2))
        matches.sort(key=lambda x: x.distance)
        matches = matches[:int(len(matches) * keep)]
        n_matches = len(matches)

        p1 = np.zeros((n_matches, 2))
        p2 = np.zeros((n_matches, 2))

        for i in range(len(matches)):
            p1[i, :] = kp1[matches[i].queryIdx].pt
            p2[i, :] = kp2[matches[i].trainIdx].pt

        tmat, inliers = cv2.estimateAffinePartial2D(p1, p2, method=cv2.RANSAC if method == 0 else cv2.LMEDS,
                                                   confidence=confidence)
        print(tmat)
        return [tmat[0][2], tmat[1][2]]



from scNodes.core.ceplugin import *
from copy import deepcopy


def create():
    return SplitRGBPlugin()


class SplitRGBPlugin(CEPlugin):
    title = "Split RGB"
    description = "Split an RGB frame into the three underlying grayscale images."

    def __init__(self):
        self.rgb_frame = None
        self.keep_original = False

    def render(self):
        _c, self.rgb_frame = self.widget_select_frame_rgb("RGB Frame:", self.rgb_frame)
        _c, self.keep_original = imgui.checkbox("Keep original", self.keep_original)
        if self.widget_centred_button("Convert"):
            rgb = self.rgb_frame.data
            r = CLEMFrame(rgb[:, :, 0])
            r.pixel_size = self.rgb_frame.pixel_size
            r.transform = deepcopy(self.rgb_frame.transform)
            r.lut = 0
            r.colour = (1.0, 0.0, 0.0, 1.0)
            r.update_lut()
            g = CLEMFrame(rgb[:, :, 1])
            g.pixel_size = self.rgb_frame.pixel_size
            g.transform = deepcopy(self.rgb_frame.transform)
            g.lut = 0
            g.colour = (0.0, 1.0, 0.0, 1.0)
            g.update_lut()
            b = CLEMFrame(rgb[:, :, 2])
            b.pixel_size = self.rgb_frame.pixel_size
            b.transform = deepcopy(self.rgb_frame.transform)
            b.lut = 0
            b.colour = (0.0, 0.0, 1.0, 1.0)
            b.update_lut()
            cfg.ce_frames.append(r)
            cfg.ce_frames.append(g)
            cfg.ce_frames.append(b)
            b.move_to_front()
            g.move_to_front()
            r.move_to_front()
            if not self.keep_original:
                cfg.correlation_editor.delete_frame(self.rgb_frame)





from scNodes.core.node import *


def create():
    return TemporalFilterNode()


class TemporalFilterNode(Node):
    title = "Temporal filter"
    group = "Image processing"
    colour = (55 / 255, 236 / 255, 54 / 255, 1.0)
    sortid = 102

    FILTERS = ["Forward difference", "Backward difference", "Central difference", "Grouped difference", "Windowed average"]
    NEGATIVE_MODES = ["Absolute", "Zero", "Retain"]
    INCOMPLETE_GROUP_MODES = ["Discard"]

    def __init__(self):
        super().__init__()  # Was: super(LoadDataNode, self).__init__()
        self.size = 250

        # Set up connectable attributes
        self.dataset_in = ConnectableAttribute(ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.INPUT, parent=self)
        self.dataset_out = ConnectableAttribute(ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.OUTPUT, parent=self)

        self.filter = 0
        self.negative_handling = 1
        self.incomplete_group_handling = 0
        self.skip = 1
        self.group_size = 11
        self.group_background_index = 1
        self.window = 3

    def render(self):
        if super().render_start():
            self.dataset_out.render_start()
            self.dataset_in.render_start()
            self.dataset_out.render_end()
            self.dataset_in.render_end()
            imgui.spacing()
            imgui.separator()
            imgui.spacing()

            imgui.push_item_width(170)
            _c, self.filter = imgui.combo("Filter", self.filter, TemporalFilterNode.FILTERS)
            self.any_change = self.any_change or _c
            imgui.pop_item_width()
            imgui.push_item_width(110)
            _c, self.negative_handling = imgui.combo("Negative handling", self.negative_handling, TemporalFilterNode.NEGATIVE_MODES)
            self.any_change = self.any_change or _c
            imgui.pop_item_width()
            imgui.push_item_width(90)
            if self.filter == 0 or self.filter == 1 or self.filter == 2:
                _c, self.skip = imgui.input_int("Step (# frames)", self.skip, 0, 0)
                self.any_change = self.any_change or _c
            elif self.filter == 3:
                _c, self.group_size = imgui.input_int("Images per cycle", self.group_size, 0, 0)
                self.any_change = self.any_change or _c
                _c, self.group_background_index = imgui.input_int("Background index", self.group_background_index, 0, 0)
                self.any_change = self.any_change or _c
                _c, self.incomplete_group_handling = imgui.combo("Incomplete groups", self.incomplete_group_handling, TemporalFilterNode.INCOMPLETE_GROUP_MODES)
                self.any_change = self.any_change or _c
            elif self.filter == 4:
                _c, self.window = imgui.input_int("Window size", self.window, 0, 0)
                self.any_change = self.any_change or _c
            imgui.pop_item_width()
            super().render_end()

    def get_image_impl(self, idx=None):
        data_source = self.dataset_in.get_incoming_node()
        if data_source:
            pxd = None
            if self.filter == 0:
                pxd = data_source.get_image(idx + self.skip).load() - data_source.get_image(idx).load()
            elif self.filter == 1:
                pxd = data_source.get_image(idx).load() - data_source.get_image(idx - self.skip).load()
            elif self.filter == 2:
                pxd = data_source.get_image(idx + self.skip).load() - data_source.get_image(idx - self.skip).load()
            elif self.filter == 3:
                pxd = data_source.get_image(idx // self.group_size).load() - data_source.get_image(self.group_size * (idx // self.group_size) + self.group_background_index).load()
            elif self.filter == 4:
                pxd = np.zeros_like(data_source.get_image(idx).load())
                for i in range(-self.window, self.window + 1):
                    pxd += data_source.get_image(idx + i).load()
                pxd /= (2 * self.window + 1)

            if self.negative_handling == 0:
                pxd = np.abs(pxd)
            elif self.negative_handling == 1:
                pxd[pxd < 0] = 0
            elif self.negative_handling == 2:
                pass

            out_image = data_source.get_image(idx)
            out_image.data = pxd
            return out_image



from scNodes.core.node import *


def create():
    return ImageCalculatorNode()


class ImageCalculatorNode(Node):
    title = "Image calculator"
    group = "Image processing"
    colour = (143 / 255, 143 / 255, 143 / 255, 1.0)
    sortid = 103
    # Note: the output dataset has all the metadata of dataset_in_a
    OPERATIONS = ["Add", "Subtract", "Divide", "Multiply"]

    def __init__(self):
        super().__init__()
        self.size = 230
        self.dataset_in = ConnectableAttribute(ConnectableAttribute.TYPE_MULTI, ConnectableAttribute.INPUT, parent=self, allowed_partner_types=[ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.TYPE_IMAGE])
        self.input_b = ConnectableAttribute(ConnectableAttribute.TYPE_MULTI, ConnectableAttribute.INPUT, parent=self, allowed_partner_types=[ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.TYPE_IMAGE])
        self.dataset_out = ConnectableAttribute(ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.OUTPUT, parent=self)
        self.image_out = ConnectableAttribute(ConnectableAttribute.TYPE_IMAGE, ConnectableAttribute.OUTPUT, parent=self)

        self.operation = 1

    def render(self):
        if super().render_start():
            self.dataset_in.render_start()
            if self.dataset_in.current_type == ConnectableAttribute.TYPE_IMAGE:
                self.image_out.render_start()
                self.image_out.render_end()
            else:
                self.dataset_out.render_start()
                self.dataset_out.render_end()
            self.dataset_in.render_end()
            imgui.spacing()
            self.input_b.render_start()
            self.input_b.render_end()

            imgui.spacing()
            imgui.separator()
            imgui.spacing()
            imgui.push_item_width(90)
            _c, self.operation = imgui.combo("Operation", self.operation, ImageCalculatorNode.OPERATIONS)
            self.any_change = self.any_change | _c
            imgui.pop_item_width()
            super().render_end()

    def get_image_impl(self, idx=None):
        try:
            source_a = self.dataset_in.get_incoming_node()
            source_b = self.input_b.get_incoming_node()
            if source_a and source_b:
                img_a = source_a.get_image(idx)
                img_b = source_b.get_image(idx)
                img_a_pxd = img_a.load()
                img_b_pxd = img_b.load()
                img_out = None
                if self.operation == 0:
                    img_out = img_a_pxd + img_b_pxd
                elif self.operation == 1:
                    img_out = img_a_pxd - img_b_pxd
                elif self.operation == 2:
                    img_out = img_a_pxd / img_b_pxd
                elif self.operation == 3:
                    img_out = img_a_pxd * img_b_pxd
                if self.dataset_in.current_type != ConnectableAttribute.TYPE_IMAGE:
                    img_a.data = img_out
                    return img_a
                else:
                    virtual_frame = Frame("virtual_frame")
                    virtual_frame.data = img_out.astype(np.uint16)
                    return virtual_frame
        except Exception as e:
            cfg.set_error(Exception(), "ImageCalculatorNode error:\n"+str(e))



"""
This module performs fourier interpolation on a single image 
or a image stack (tiff file). Interpolated image (stack) can be 
saved into a new tiff file or return as a numpy array.

Fourier interpolation was implemented with transformation matrix 
operation, where the Fourier transformation matrix was constructed 
with the original matrix size without interpolation grids, and the 
inverse Fourier transformation matrix encodes the extra interpolation
position coordinates.

Adapted from https://github.com/xiyuyi/xy_fInterp_forTIFF.
"""

import numpy as np

try:
    from . import switches as s
except:
    import switches as s

if s.SPHINX_SWITCH is False:
    import tifffile as tiff

import sys


def _base_vect_generator2d(xrange, yrange):
    """Generate base vectors for x- and y-dimension."""
    bx, by = np.zeros(xrange), np.zeros(yrange)
    bx[1], by[1] = 1, 1
    bx, by = np.fft.fft(bx), np.fft.fft(by)
    # bx, by = fftshift(fft(ifftshift(bx))), fftshift(fft(ifftshift(by)))
    return bx, by


def _calc_ft_matrix(base, spectrum_range):
    """
    Calculate fourier transform matrix without center shift from base vectors.
    """
    power_matrix = np.ones((spectrum_range, spectrum_range))
    power_matrix = np.arange(spectrum_range).reshape(spectrum_range, 1) * \
                   power_matrix
    ft_matrix = np.power(base, power_matrix)
    return ft_matrix


def _calc_ift_matrix(base, spectrum_range, interp_num):
    """
    Calculate inverse fourier transform matrix without center shift from base
    vectors for a given fold of interpolation.
    """
    conj_base = np.reshape(np.conj(base), (1, spectrum_range))
    ift = np.matmul(np.ones(((spectrum_range - 1) * interp_num + 1, 1)),
                    conj_base)
    iftp = np.arange(0, spectrum_range - 1 + 1e-10, 1 / interp_num)
    iftp = np.matmul(iftp.reshape(-1, 1), np.ones((1, spectrum_range)))
    ift = np.power(ift, iftp) / spectrum_range
    return ift


def ft_matrix2d(xrange, yrange):
    """Calculate fourier transform matrix for x- and y-dimension."""
    bx, by = _base_vect_generator2d(xrange, yrange)
    fx = _calc_ft_matrix(bx, xrange)
    fy = _calc_ft_matrix(by, yrange).T
    return fx, fy


def ift_matrix2d(xrange, yrange, interp_num):
    """
    Calculate inverse fourier transform matrix without center shift for a 
    given fold of interpolation for x- and y-dimension. Here we use 2 times 
    the x-dimension of a frame image for xrange and 2 times of y for yrange. 
    interp_num is the number of times the resolution enhanced. For instance, 
    when interp_num = 2, the resolution is two times the original one.
    """
    bx, by = _base_vect_generator2d(xrange, yrange)
    ifx = _calc_ift_matrix(bx, xrange, interp_num)
    ify = _calc_ift_matrix(by, yrange, interp_num).T

    return ifx, ify


def interpolate_image(im, fx, fy, ifx, ify, interp_num):
    """
    Performs fourier interpolation to increase the resolution of the image. 
    The interpolated image can be further processed by SOFI for 
    super-resolution imaging.

    Parameters
    ----------
    im : ndarray
        Input image.
    fx : ndarray
        Fourier transform matrix in x generated from ft_matrix2d function.
    fy : ndarray
        Fourier transform matrix in y generated from ft_matrix2d function.
    ifx : ndarray
        Inverse fourier transform matrix in x from ift_matrix2d function.
    ify : ndarray
        Inverse fourier transform matrix in y from ift_matrix2d function.
    interp_num : int
        The number of pixels to be interpolated between two adjacent pixels.

    Returns
    -------
    interp_im : ndarray
        Interpolated image with new dimensions.

    Notes
    -----
    Please note here that interpolation doesn' extend over the edgeo f the 
    matrix, therefore the total number of pixels in the resulting matrix is 
    not an integer fold of the original size of the matrix. For example, if 
    the original matrix size of each frame is xdim and ydim for x and y 
    dimensions respectively. After interpolation, the size of the resulting 
    matrix will be ((xdim-1)*f + 1) for x dimension, and ((ydim-1)*f + 1) 
    for y dimension.

    Example
    -------
    ::

        import numpy as np
        import matplotlib.pyplot as plt
        xdim, ydim, sigma, mu = 10, 10, 0.5, 0
        x, y = np.meshgrid(np.linspace(-1, 1, xdim), np.linspace(-1, 1, xdim))
        im = np.exp(-((np.sqrt(x*x + y*y) - mu)**2 / (2*sigma**2)))
        xrange, yrange, interp_num = 2 * xdim, 2 * ydim, 3
        fx, fy = ft_matrix2d(xrange, yrange)
        ifx, ify = ift_matrix2d(xrange, yrange, interp_num)
        interp_im = interpolate_image(im, fx, fy, ifx, ify, interp_num)

    """
    xdim, ydim = np.shape(im)

    # 1. Extend im to create the natural peoriocity in the resulting image 
    # to avoid ringing artifacts after fourier interpolation.
    ext_im = np.append(np.append(im, np.fliplr(im), axis=1),
                       np.append(np.flipud(im), np.rot90(im, 2),
                                 axis=1), axis=0)

    # 2. Fourier transform
    fall = np.matmul(np.matmul(fx, ext_im), fy)
    # fall=fx@ext_im@fy for Python 3.5 or newer versions

    # 3. Inverse Fourier transform
    ifall = np.absolute(np.dot(np.dot(ifx, fall), ify))

    # 4. Take the region corresponding to the FOV of the original image
    xdim_new = (xdim - 1) * interp_num + 1
    ydim_new = (ydim - 1) * interp_num + 1
    interp_im = ifall[:xdim_new, :ydim_new]

    return interp_im


def fourier_interp_array(im, interp_num_lst):
    """
    Performs fourier interpolation on an image array with a list of 
    interpolation factors.
    """
    xdim, ydim = im.shape
    xrange, yrange = 2 * xdim, 2 * ydim  # define the ft spectrum span
    fx, fy = ft_matrix2d(xrange, yrange)
    interp_im_lst = []
    for interp_num in interp_num_lst:
        ifx, ify = ift_matrix2d(xrange, yrange, interp_num)
        interp_im = interpolate_image(im, fx, fy, ifx, ify, interp_num)
        # interp_im_lst.append(np.int_(np.around(interp_im)))
        interp_im_lst.append(interp_im)

    return interp_im_lst


def fourier_interp_tiff(filepath, filename, interp_num_lst, frames=[],
                        save_option=True, return_option=False):
    """
    Performs fourier interpolation on a tiff image (stack) with a list of 
    interpolation factors (the number of pixels to be interpolated between 
    two adjacent pixels).

    Parameters
    ----------
    filepath : str
        The path to the tiff file.
    filename : str
        The filename of the tiff file without '.tif'.
    interp_num_lst : list (int)
        A list of interpolation factors.
    frames : list of int
        The start and end frame number.
    save_option : bool
        Whether to save the interpolated images into tiff files (each 
        interpolation factor seperately).
    return_option : bool
        Whether to return the interpolated image series as 3d arrays.

    Returns
    -------
    interp_imstack_lst : list (ndarray)
        A list of interpolated image series corresponding to the interpolation
        factor list.
    """
    imstack = tiff.TiffFile(filepath + '/' + filename + '.tif')
    xdim, ydim = np.shape(imstack.pages[0])
    xrange, yrange = 2 * xdim, 2 * ydim
    fx, fy = ft_matrix2d(xrange, yrange)
    interp_imstack_lst = []

    # if user did not select the video length, process on the whole video
    if frames:
        for interp_num in interp_num_lst:
            ifx, ify = ift_matrix2d(xrange, yrange, interp_num)
            interp_imstack = []
            print('Calculating interpolation factor = %d...' % interp_num)
            for frame_num in range(frames[0], frames[1]):
                # read a frame from the original tiff file
                im = tiff.imread(filepath + '/' + filename + '.tif', key=frame_num)
                # find the minimum and maximum values of this image
                immax = np.max(im.ravel())
                immin = np.min(im.ravel())
                # perform interpolation
                interp_im = interpolate_image(im, fx, fy, ifx, ify, interp_num)
                # ensure the interpolated image have the identical dynamic range of the original image
                interp_immax = np.max(interp_im.ravel())
                interp_immin = np.min(interp_im.ravel())
                interp_im = (interp_im - interp_immin) / (interp_immax - interp_immin)
                interp_im = interp_im * (immax - immin) + immin

                #                interp_im = np.int_(np.around(interp_im))
                interp_im = np.uint16(np.around(interp_im))
                sys.stdout.write('\r')
                sys.stdout.write("[{:{}}] {:.1f}%".format(
                    "=" * int(30 / (frames[1] - frames[0]) * (frame_num - frames[0] + 1)), 29,
                    (100 / (frames[1] - frames[0]) * (frame_num - frames[0] + 1))))
                sys.stdout.flush()
                if save_option is True:
                    tiff.imwrite(filepath + '/' + filename + '_InterpNum' + str(interp_num) +
                                 '.tif', interp_im, dtype='uint16', append=True)
                if return_option is True:
                    interp_imstack.append(interp_im)

            if return_option is True:
                interp_imstack_lst.append(interp_imstack)
            print('\n')
    else:
        mvlength = len(imstack.pages)
        for interp_num in interp_num_lst:
            ifx, ify = ift_matrix2d(xrange, yrange, interp_num)
            interp_imstack = []
            print('Calculating interpolation factor = %d...' % interp_num)
            for frame_num in range(mvlength):
                # read a frame from the original tiff file
                im = tiff.imread(filepath + '/' + filename + '.tif', key=frame_num)
                # find the minimum and maximum values of this image
                immax = np.max(im.ravel())
                immin = np.min(im.ravel())
                # perform interpolation
                interp_im = interpolate_image(im, fx, fy, ifx, ify, interp_num)
                # ensure the interpolated image have the identical dynamic range of the original image
                interp_immax = np.max(interp_im.ravel())
                interp_immin = np.min(interp_im.ravel())
                interp_im = (interp_im - interp_immin) / (interp_immax - interp_immin)
                interp_im = interp_im * (immax - immin) + immin

                #                interp_im = np.int_(np.around(interp_im))
                interp_im = np.uint16(np.around(interp_im))
                sys.stdout.write('\r')
                sys.stdout.write("[{:{}}] {:.1f}%".format(
                    "=" * int(30 / mvlength * (frame_num + 1)), 29,
                    (100 / mvlength * (frame_num + 1))))
                sys.stdout.flush()
                if save_option is True:
                    tiff.imwrite(filepath + '/' + filename + '_InterpNum' + str(interp_num) +
                                 '.tif', interp_im, dtype='uint16', append=True)
                if return_option is True:
                    interp_imstack.append(interp_im)

            if return_option is True:
                interp_imstack_lst.append(interp_imstack)
            print('\n')

    if return_option is True:
        return interp_imstack_lst


from scNodes.core.opengl_classes import *
from scNodes.core.dataset import *
from scNodes.core.util import *
import scNodes.core.settings as settings


class Reconstructor:
    default_pixel_size = 6.4
    default_uncertainty = 100.0 / 6.4 * default_pixel_size
    kernel_size = 128
    default_img_size = (5000, 5000)
    default_tile_size = (2048, 2048)
    vertices = [-0.5, 0.5, 0.0, 1.0, -0.5, -0.5, 0.0, 0.0, 0.5, -0.5, 1.0, 0.0, 0.5, 0.5, 1.0, 1.0]
    indices = [0, 1, 2, 2, 0, 3]

    def __init__(self):
        # Shader
        self.shader = Shader(cfg.root+"shaders/ne_reconstructor_shader.glsl")
        # create texture, gaussian spot img, set pixel sizes etc.
        self.texture = Texture(format="r32f")
        self.texture.set_linear_interpolation()

        self.pixel_size = Reconstructor.default_pixel_size  # nm per pixel
        self.quad_uncertainty = Reconstructor.default_uncertainty
        self.quad_size = Reconstructor.kernel_size
        self.image_size = Reconstructor.default_img_size
        self.tile_size = Reconstructor.default_tile_size
        self.set_pixel_size(Reconstructor.default_pixel_size)
        self.fbo = FrameBuffer(*self.tile_size, texture_format="rgb32f")
        self.vao = VertexArray(VertexBuffer(Reconstructor.vertices), IndexBuffer(Reconstructor.indices),
                               attribute_format="xyuv")
        self.lut_texture = Texture(format="rgb32f")
        self.set_lut(0)
        self.instance_vbos = dict()
        self.particle_data = ParticleData()
        self.camera = StaticCamera(Reconstructor.default_tile_size)

        self.mode = "ui16"
        self.camera_origin = [0.0, 0.0]
        self.colours_set = False

    def set_lut(self, lut_idx):
        lut_array = np.asarray(settings.luts[settings.lut_names[lut_idx]])
        if lut_array.shape[1] == 3:
            lut_array = np.reshape(lut_array, (1, lut_array.shape[0], 3))
            self.lut_texture.update(lut_array)

    def set_mode(self, mode="float"):
        """
        :param mode: either "float" or "ui16." Save as 32 bit float for best quality, or as 16 bit unsigned int when
        using the render result in ImageViewer (rgb32f reconstructions quickly exceed GPU memory).
        """
        if self.mode in ["float", "ui16"]:
            self.mode = mode

    def set_pixel_size(self, pixel_size):
        """
        :param pixel_size: float, pixel size in nm for the final reconstruction.
        """
        self.pixel_size = pixel_size

        # Reset the default-gaussian-quad to use the new pixel size.
        default_sigma = self.default_uncertainty / self.pixel_size
        _g_img_axes = np.linspace(-(Reconstructor.kernel_size - 1) / 2., (Reconstructor.kernel_size - 1) / 2.,
                                  Reconstructor.kernel_size)
        _g_img_gauss = np.exp(-0.5 * np.square(_g_img_axes) / np.square(default_sigma))
        _g_img_kernel = np.outer(_g_img_gauss, _g_img_gauss)
        default_gaussian = _g_img_kernel / np.sum(_g_img_kernel)
        self.texture.update(default_gaussian)

    def set_image_size(self, image_size):
        self.image_size = [image_size[0] // 2 * 2, image_size[1] // 2 * 2]

    def set_particle_data(self, particle_data):
        """
        :param particle_data: object of type ParticleData. Sets the particle dataset for the reconstructor to render.
        :return:
        """
        self.particle_data = particle_data

    def update_particle_data(self):
        self.create_instance_buffers()
        self.particle_data.baked_by_renderer = True

    def recompile_shader(self):
        """
        Useful during debugging of the shaders.
        :return:
        """
        try:
            shader = Shader("shaders/ne_reconstructor_shader.glsl")
            self.shader = shader
        except Exception as e:
            raise e

    def render(self, fixed_uncertainty=None):
        """
        Render the particles in the currently st ParticleData obj. (see set_particle_data()), at the current pixel size (see set_pixel_size()).
        :return: a 3D [width, height, 3] numpy array of type float (when in mode 'float') or uint16 (mode 'ui16'). See set_mode()
        """
        print("Start render in reconstruction.py")
        if not self.particle_data.baked_by_renderer:
            self.update_particle_data()  # refresh the instance buffers

        if not self.colours_set:
            self.update_particle_colours()
        if fixed_uncertainty is not None:
            self.set_fixed_uncertainty_value(fixed_uncertainty)

        self.update_particle_colours()
        self.update_particle_states()
        glEnable(GL_BLEND)
        glBlendEquation(GL_FUNC_ADD)
        glBlendFunc(GL_ONE, GL_ONE)
        self.fbo.clear((0.0, 0.0, 0.0, 1.0))
        self.fbo.bind()
        self.vao.bind()
        self.shader.bind()
        self.shader.uniform1f("quad_pixel_size", Reconstructor.kernel_size)
        self.shader.uniform1f("quad_uncertainty", self.quad_uncertainty)
        self.shader.uniform1f("pixel_size", self.pixel_size)
        self.texture.bind(0)
        self.lut_texture.bind(1)
        # Make empty image
        W = self.image_size[0]
        w = self.tile_size[0]
        H = self.image_size[1]
        h = self.tile_size[1]

        sr_image = np.zeros((H, W, 3), dtype=np.float32)
        tiles_x = int(np.ceil(W / w))
        tiles_y = int(np.ceil(H / h))
        for i in range(tiles_x):
            for j in range(tiles_y):
                tile = self.render_tile((i, j))
                sr_image[j * h:min([(j + 1) * h, H]), i * w:min([(i + 1) * w, W]), :] = tile[:min([h, H - j * h]), :min([w, W - i * w]), :]
        self.shader.unbind()
        self.vao.unbind()
        self.fbo.unbind()
        if fixed_uncertainty is not None:
            self.undo_fixed_uncertainty_value()
        if self.mode == "float":
            return sr_image
        elif self.mode == "ui16":
            sr_image_ui16 = np.zeros((H, W, 3), dtype=np.uint16)
            _normfac = np.amax(sr_image)
            if _normfac == 0.0:
                _normfac = 1
            r = sr_image[:, :, 0] / _normfac * 65535
            g = sr_image[:, :, 1] / _normfac * 65535
            b = sr_image[:, :, 2] / _normfac * 65535
            sr_image_ui16[:, :, 0] = r.astype(np.uint16)
            sr_image_ui16[:, :, 1] = g.astype(np.uint16)
            sr_image_ui16[:, :, 2] = b.astype(np.uint16)
            return sr_image_ui16

    def set_camera_origin(self, camera_origin):
        """
        :param camera_origin: origin position in world coordinates for the camera.
        Super-resolution images are rendered in tiles, since the full output image size is often larger than
        GPU memory constraints would efficiently allow. The nr. of tiles if automatically determined in render()
        Prior to rendering a reconstruction, the camera origin must be set to the top-left coordinate of the
        reconstruction ROI.
        :return:
        """
        self.camera_origin = camera_origin

    def create_instance_buffers(self):
        """
        Generate and populate the instance buffers: GPU-side data arrays of particle position (x, y), uncertainty, colour, etc.
        :return:
        """
        self.vao.bind()

        self.instance_vbos["x"] = VertexBuffer(self.particle_data.parameter['x [nm]'])
        self.instance_vbos["x"].set_location_and_stride(2, 1)
        self.instance_vbos["x"].set_divisor_to_per_instance()

        self.instance_vbos["y"] = VertexBuffer(self.particle_data.parameter['y [nm]'])
        self.instance_vbos["y"].set_location_and_stride(3, 1)
        self.instance_vbos["y"].set_divisor_to_per_instance()

        print(self.particle_data.parameter['uncertainty [nm]'])
        self.instance_vbos["uncertainty"] = VertexBuffer(self.particle_data.parameter['uncertainty [nm]'])
        self.instance_vbos["uncertainty"].set_location_and_stride(4, 1)
        self.instance_vbos["uncertainty"].set_divisor_to_per_instance()

        self.instance_vbos["colour_idx"] = VertexBuffer()
        self.instance_vbos["colour_idx"].set_location_and_stride(5, 1)
        self.instance_vbos["colour_idx"].set_divisor_to_per_instance()

        self.instance_vbos["state"] = VertexBuffer()
        self.instance_vbos["state"].set_location_and_stride(6, 1)
        self.instance_vbos["state"].set_divisor_to_per_instance()
        self.vao.unbind()

        self.update_particle_colours()

    def set_fixed_uncertainty_value(self, value):
        self.vao.bind()
        self.instance_vbos["uncertainty"].update(np.ones((self.particle_data.n_particles, 1)) * value)
        self.vao.unbind()

    def undo_fixed_uncertainty_value(self):
        self.vao.bind()
        self.instance_vbos["uncertainty"].update(self.particle_data.parameter['uncertainty [nm]'])
        self.vao.unbind()

    def update_particle_colours(self):
        self.vao.bind()
        _colours_idx = list()
        for particle in self.particle_data.particles:
            _colours_idx.append(particle.colour_idx)
        self.instance_vbos["colour_idx"].update(np.asarray(_colours_idx).flatten(order = "C"))
        self.vao.unbind()
        self.colours_set = True

    def update_particle_states(self):
        self.vao.bind()
        _states = list()
        for particle in self.particle_data.particles:
            _states.append(particle.visible * 1.0)
        self.instance_vbos["state"].update(np.asarray(_states).flatten(order = "C"))
        self.vao.unbind()

    def render_tile(self, tile_idx=(0, 0)):
        self.fbo.clear((0.0, 0.0, 0.0, 1.0))
        self.fbo.bind()
        camera_x = -(tile_idx[0] + 0.5) * self.tile_size[0]
        camera_y = -(tile_idx[1] + 0.5) * self.tile_size[1]

        print("Tile size is ", self.tile_size)
        self.camera.position = [camera_x + self.camera_origin[0], camera_y + self.camera_origin[1], 0.0]
        print("Camera position x, y is: ", self.camera.position[0], self.camera.position[1])
        self.camera.update_matrix()
        self.shader.uniformmat4("cameraMatrix", self.camera.view_projection_matrix)
        glDrawElementsInstanced(GL_TRIANGLES, self.vao.indexBuffer.getCount(), GL_UNSIGNED_SHORT, None,
                                self.particle_data.n_particles)
        print("Done rendering tile")
        tile = glReadPixels(0, 0, self.default_tile_size[0], self.default_tile_size[1], GL_RGB, GL_FLOAT)
        print("Done reading tile")
        return tile


class StaticCamera:
    def __init__(self, sensor_size):
        self.position = np.asarray([0.0, 0.0, 0.0])
        self.zoom = 1.0
        self.view_matrix = np.identity(4)
        self.projection_matrix = np.identity(4)
        self.view_projection_matrix = np.identity(4)
        self.sensor_size = sensor_size
        self.update_matrix()

    def update_matrix(self):
        self.projection_matrix = np.matrix([
            [2 / self.sensor_size[0], 0, 0, 0],
            [0, 2 / self.sensor_size[1], 0, 0],
            [0, 0, -2 / 100, 0],
            [0, 0, 0, 1],
        ])
        self.view_matrix = np.matrix([
            [self.zoom, 0.0, 0.0, self.position[0]],
            [0.0, self.zoom, 0.0, self.position[1]],
            [0.0, 0.0, self.zoom, self.position[2]],
            [0.0, 0.0, 0.0, 1.0],
        ])
        self.view_projection_matrix = np.matmul(self.projection_matrix, self.view_matrix)



from scNodes.core.node import *
import cv2


def create():
    return ScaleImageNode()


class ScaleImageNode(Node):
    title = "Scale image"
    group = "Image processing"
    colour = (143 / 255, 143 / 255, 143 / 255, 1.0)
    sortid = 110
    CV2_INTERPOLATION_METHODS = [cv2.INTER_NEAREST, cv2.INTER_LINEAR, cv2.INTER_AREA, cv2.INTER_CUBIC]

    def __init__(self):
        super().__init__()
        self.size = 170
        self.dataset_in = ConnectableAttribute(ConnectableAttribute.TYPE_MULTI, ConnectableAttribute.INPUT, self, [ConnectableAttribute.TYPE_IMAGE, ConnectableAttribute.TYPE_DATASET])
        self.dataset_in.colour = ConnectableAttribute.COLOUR[ConnectableAttribute.TYPE_DATASET]
        self.dataset_out = ConnectableAttribute(ConnectableAttribute.TYPE_MULTI, ConnectableAttribute.OUTPUT, self, [ConnectableAttribute.TYPE_IMAGE, ConnectableAttribute.TYPE_DATASET])
        self.dataset_out.colour = ConnectableAttribute.COLOUR[ConnectableAttribute.TYPE_DATASET]
        self.fac_x = 0.60
        self.fac_y = 0.60
        self.interpolation = 0

    def render(self):
        if super().render_start():
            self.dataset_in.render_start()
            self.dataset_out.render_start()
            self.dataset_in.render_end()
            self.dataset_out.render_end()

            imgui.spacing()
            imgui.separator()
            imgui.spacing()

            imgui.text("Scale factors:")
            imgui.push_item_width(40)
            _c, self.fac_x = imgui.input_float("x", self.fac_x, format="%.2f")
            self.any_change = _c or self.any_change
            if _c:
                self.fac_y = self.fac_x
            _c, self.fac_y = imgui.input_float("y", self.fac_y, format="%.2f")
            self.any_change = _c or self.any_change
            imgui.pop_item_width()
            imgui.spacing()
            imgui.text("Interpolation:")
            imgui.set_next_item_width(150)
            _c, self.interpolation = imgui.combo("##interp", self.interpolation, ["Nearest neighbour", "Linear", "cv2 INTER_AREA", "Cubic"])
            self.any_change = _c or self.any_change
            super().render_end()

    def get_image_impl(self, idx=None):
        data_source = self.dataset_in.get_incoming_node()
        if data_source:
            outframe = data_source.get_image(idx).clone()
            pxd = outframe.load()
            outsize = (int(np.shape(pxd)[0] * self.fac_x), int(np.shape(pxd)[1] * self.fac_y))
            outframe.data = cv2.resize(pxd, outsize, self.fac_x, self.fac_y, ScaleImageNode.CV2_INTERPOLATION_METHODS[self.interpolation])
            outframe.pixel_size /= self.fac_x
            outframe.width, outframe.height = np.shape(outframe.data)[0:2]
            return outframe




from scNodes.core.node import *
import os
from tkinter import filedialog
import shutil
from joblib.externals.loky import set_loky_pickler
set_loky_pickler("dill")


def create():
    return BakeStackNode()


class BakeStackNode(Node):
    title = "Bake stack"
    group = "Image processing"
    colour = (143 / 255, 143 / 255, 143 / 255, 1.0)
    sortid = 120

    RANGE_OPTIONS = ["All frames", "Custom range"]

    ## todo fix error where bakestack freezes upon leaving entire node setup intact, changing source data node input to a different stack, then pressing GO on bake stack node.
    def __init__(self):
        super().__init__()
        self.size = 230

        self.dataset_in = ConnectableAttribute(ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.INPUT, self)
        self.dataset_out = ConnectableAttribute(ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.OUTPUT, self)
        self.coordinates_in = ConnectableAttribute(ConnectableAttribute.TYPE_COORDINATES, ConnectableAttribute.INPUT, self)
        self.coordinates_out = ConnectableAttribute(ConnectableAttribute.TYPE_COORDINATES, ConnectableAttribute.OUTPUT, self)

        self.parallel = True
        self.baking = False
        self.n_to_bake = 1
        self.n_baked = 0
        self.range_option = 0
        self.custom_range_min = 0
        self.custom_range_max = 1
        self.has_dataset = False
        self.dataset = Dataset()
        self.frames_to_bake = list()
        self.temp_dir_do_not_edit = "_srnodes_temp_dir_bakestack_"+str(self.id)
        self.joblib_dir_counter = 0
        self.baked_at = 0
        self.has_coordinates = False
        self.coordinates = list()
        self.bake_coordinates = False

    def render(self):
        if super().render_start():
            self.dataset_in.render_start()
            self.dataset_in.render_end()
            if self.has_dataset:
                self.dataset_out.render_start()
                self.dataset_out.render_end()
            if self.bake_coordinates:
                imgui.spacing()
                self.coordinates_in.render_start()
                self.coordinates_in.render_end()
                if self.has_coordinates:
                    self.coordinates_out.render_start()
                    self.coordinates_out.render_end()

            imgui.spacing()
            imgui.separator()
            imgui.spacing()

            _c, self.bake_coordinates = imgui.checkbox("Also bake coordinates", self.bake_coordinates)
            if _c and not self.bake_coordinates:
                self.coordinates_in.disconnect_all()

            if self.range_option == 1:
                imgui.push_item_width(80)
                _c, (self.custom_range_min, self.custom_range_max) = imgui.input_int2('[start, top) index', self.custom_range_min, self.custom_range_max)
                imgui.pop_item_width()
            _c, self.parallel = imgui.checkbox("Parallel processing", self.parallel)
            imgui.set_next_item_width(100)
            _c, self.range_option = imgui.combo("Range to bake", self.range_option, BakeStackNode.RANGE_OPTIONS)
            clicked, self.baking = self.play_button()
            if clicked and self.baking:
                self.init_bake()

            if self.baking:
                imgui.text("Baking process:")
                self.progress_bar(self.n_baked / self.n_to_bake)
                imgui.spacing()
                imgui.spacing()
                imgui.spacing()

            if self.has_dataset:
                imgui.text(f"Output data was baked at: {self.baked_at}")
                header_expanded, _ = imgui.collapsing_header("Advanced", None)
                if header_expanded:
                    self.render_advanced()
            super().render_end()


    def render_advanced(self):
        _cw = imgui.get_content_region_available_width()
        imgui.new_line()
        imgui.same_line(spacing = _cw / 2 - 80 / 2)
        if imgui.button("Quick save", 80, 25):
            fpath = filedialog.asksaveasfilename()
            if fpath is not None:
                if '.' in fpath[-5:]:
                    fpath = fpath[:fpath.rfind(".")]
            # now copy the files from the temp folder to the requested folder
            if not os.path.isdir(fpath):
                os.mkdir(fpath)
            imglist = glob.glob(self.temp_dir_do_not_edit +"/*.tif*")
            for img in imglist:
                shutil.copy(img, fpath)

    def get_image_and_save(self, idx=None):
        datasource = self.dataset_in.get_incoming_node()
        if datasource:
            pxd = datasource.get_image(idx).load()
            Image.fromarray(pxd).save(self.temp_dir_do_not_edit + "/0" + str(idx) + ".tif")
            if self.bake_coordinates:
                coordsource = self.coordinates_in.get_incoming_node()
                coordinates = coordsource.get_coordinates(idx)
                return coordinates
            return False
        return False

    def get_image_impl(self, idx=None):
        if self.has_dataset and idx in range(0, self.dataset.n_frames):
            retimg = copy.deepcopy(self.dataset.get_indexed_image(idx))
            retimg.clean()
            if self.has_coordinates:
                retimg.maxima = self.coordinates[idx]
            return retimg
        else:
            datasource = self.dataset_in.get_incoming_node()
            if datasource:
                return datasource.get_image(idx)

    def get_coordinates(self, idx=None):
        if self.has_coordinates and idx in range(0, self.dataset.n_frames):
            return self.coordinates[idx]

    def init_bake(self):
        self.dataset = None
        if os.path.isdir(self.temp_dir_do_not_edit):
            for f in glob.glob(self.temp_dir_do_not_edit+"/*"):
                os.remove(f)
        else:
            os.mkdir(self.temp_dir_do_not_edit)
        self.has_dataset = False
        self.has_coordinates = False
        self.coordinates = list()
        if self.range_option == 0:
            dataset_source = Node.get_source_load_data_node(self)
            self.frames_to_bake = list(range(0, dataset_source.dataset.n_frames))
        elif self.range_option == 1:
            self.frames_to_bake = list(range(self.custom_range_min, self.custom_range_max))
        self.n_baked = 0
        self.baked_at = datetime.datetime.now().strftime("%H:%M")
        self.n_to_bake = max([1, len(self.frames_to_bake)])

    def on_update(self):
        if self.baking:
            if cfg.profiling:
                time_start = time.time()
            try:
                if self.parallel:
                    indices = list()
                    for i in range(min([cfg.batch_size, len(self.frames_to_bake)])):
                        self.n_baked += 1
                        indices.append(self.frames_to_bake[-1])
                        self.frames_to_bake.pop()
                    coordinates = Parallel(n_jobs=cfg.batch_size, mmap_mode=settings.joblib_mmmode)(delayed(self.get_image_and_save)(index) for index in indices)
                else:
                    index = self.frames_to_bake[-1]
                    self.frames_to_bake.pop()
                    coordinates = self.get_image_and_save(index)
                    self.n_baked += 1
                    ## todo fix error where bakestack freezes upon leaving entire node setup intact, changing source data node input to a different stack, then pressing GO on bake stack node.
                if self.bake_coordinates:
                    self.coordinates += coordinates  # coordinates is a list of lists. in the end, self.coordinates will be a list of length [amount of frames], with a sublist of xy coords for every img.
                if cfg.profiling:
                    self.profiler_time += time.time() - time_start
                    self.profiler_count += len(indices)

                if len(self.frames_to_bake) == 0:
                    # Load dataset from temp dir.
                    self.dataset = Dataset(self.temp_dir_do_not_edit + "/00.tif")
                    print("Done baking - now loading from disk")
                    for frame in self.dataset.frames:
                        frame.load()
                    print("Done loading")
                    self.any_change = True
                    self.baking = False
                    self.play = False
                    self.has_dataset = True
                    if self.bake_coordinates:
                        self.coordinates.reverse()
                        self.has_coordinates = True
            except Exception as e:
                self.baking = False
                self.play = False
                cfg.set_error(e, "Error baking stack: \n"+str(e))

    def pre_save_impl(self):
        cfg.pickle_temp["dataset"] = self.dataset
        self.dataset = Dataset()

    def post_save_impl(self):
        self.dataset = cfg.pickle_temp["dataset"]



import glfw
import imgui
from scNodes.core import settings
from scNodes.core import config as cfg
from scNodes.core.dataset import *
from scNodes.core.roi import *
from tkinter import filedialog
from scNodes.core.util import save_tiff


class ImageViewer:
    NO_IMAGE_NOTIFICATION_SIZE = (300, 20)
    MARKER_SIZE = 4.5
    MARKER_COLOUR = (230 / 255, 174 / 255, 13 / 255, 1.0)

    COLOUR_FRAME_BACKGROUND = (0.24, 0.24, 0.24)
    COLOUR_MAIN = (0.3 / 1.3, 0.3 / 1.3, 0.3 / 1.3)
    COLOUR_THEME_ACTIVE = (0.59, 1.0, 0.86)
    COLOUR_THEME = (0.729, 1.0, 0.95)
    COLOUR_MAIN_BRIGHT = (0.3 / 1, 0.3 / 1, 0.3 / 1)
    COLOUR_MAIN_DARK = (0.3 / 1.5, 0.3 / 1.5, 0.3 / 1.5)
    COLOUR_WINDOW_BACKGROUND = (0.14, 0.14, 0.14, 0.8)
    COLOUR_CLEAR = (0.94, 0.94, 0.94, 1.0)

    CONTRAST_WINDOW_SIZE = [200, 240]
    CONTEXT_MENU_SIZE = [200, 100]
    INFO_BAR_HEIGHT = 100

    CAMERA_PAN_SPEED = 1.0
    CAMERA_ZOOM_MAX = 50
    CAMERA_ZOOM_STEP = 0.1

    HISTOGRAM_BINS = 50
    AUTOCONTRAST_SUBSAMPLE = 2
    AUTOCONTRAST_SATURATE = 1

    FRAME_SELECT_BUTTON_SIZE = [20, 19]
    FRAME_SELECT_BUTTON_SPACING = 4

    def __init__(self, window, imgui_context, imgui_implementation):
        # glfw and imgui
        self.window = window
        self.window.clear_color = ImageViewer.COLOUR_CLEAR
        self.window.make_current()

        self.imgui_context = imgui_context
        self.imgui_implementation = imgui_implementation

        # Rendering related objects and vars
        self.shader = Shader(cfg.root+"shaders/textured_shader.glsl")
        self.roi_shader = Shader(cfg.root+"shaders/line_shader.glsl")
        self.texture = Texture(format="rgb32f")
        self.fbo = FrameBuffer(*settings.def_img_size)
        self.va = VertexArray()
        self.lut_texture = Texture(format="rgb32f")
        self.current_lut = -1
        self.base_lut = 0
        self.node_specific_lut_in_use = False
        self.lut_array = None
        self.set_lut(0)
        self.camera = Camera()

        # GUI vars
        self.autocontrast = [True, True, True]
        self.contrast_min = [0, 0, 0]
        self.contrast_max = [65535, 65535, 65535]

        self.contrast_window_open = False
        self.contrast_window_channel = 0
        self.contrast_window_position = [210, 240]
        self.context_menu_open = False
        self.context_menu_can_close = False
        self.context_menu_position = [0, 0]

        # Change flags
        self.image_size_changed = False
        self.image_requires_update = False
        self.previous_active_node = None
        self.new_image_requested = False
        # Image data
        self.image = None
        self.image_pxd = None
        self.image_width = None
        self.image_height = None
        self.image_amax = [0, 0, 0]
        self.image_amin = [0, 0, 0]
        self.hist_counts = [list(), list(), list()]
        self.hist_bins = [list(), list(), list()]
        self.mode = "R"
        self.set_image(np.zeros((16, 16)))
        self.show_image = False

        # ROI data - to be made compatible with NodeEditor later
        self.drawing_roi = False
        self.moving_roi = False
        self.roi = ROI([0, 0, 0, 0], (1.0, 0.0, 1.0, 1.0))
        self.marker = Marker([-ImageViewer.MARKER_SIZE, 0.0, ImageViewer.MARKER_SIZE, 0.0, 0.0, -ImageViewer.MARKER_SIZE, 0.0, ImageViewer.MARKER_SIZE], [0, 1, 2, 3], ImageViewer.MARKER_COLOUR)

        # GUI behaviour
        self.show_frame_select_window = True
        self.node_blocking_frame_window = False
        #
        self.current_dataset = Dataset()
        self.frame_info = ""

    def set_mode(self, mode):
        if mode in ["R", "RGB"]:
            self.mode = mode
        else:
            print(f"ImageViewer.set_mode with mode = {mode} is not a valid mode!")

    def on_update(self):
        imgui.set_current_context(self.imgui_context)
        self.window.make_current()
        self.window.set_full_viewport()
        if self.window.focused:
            self.imgui_implementation.process_inputs()
        if self.window.window_gained_focus:
            self.window.pop_any_mouse_event()

        self.window.on_update()
        self.imgui_implementation.refresh_font_texture()
        if self.window.window_size_changed:
            cfg.iv_window_height = self.window.height
            cfg.iv_window_width = self.window.width
            self.camera.set_projection_matrix(self.window.width, self.window.height)
            self.window.window_size_changed = False

        self.new_image_requested = False
        if self.image_requires_update:
            self.update_image()
            self.image_requires_update = False

        if self.image_size_changed:
            self.image_size_changed = False
            self.center_camera()

        self.camera.on_update()
        self._camera_control()
        self._render()
        self._edit_and_render_roi()

        # imgui
        imgui.get_io().display_size = self.window.width, self.window.height
        imgui.new_frame()
        # Push overall imgui style vars
        imgui.push_style_color(imgui.COLOR_WINDOW_BACKGROUND, *cfg.COLOUR_PANEL_BACKGROUND)
        imgui.push_style_color(imgui.COLOR_TEXT, *cfg.COLOUR_TEXT)
        imgui.push_style_color(imgui.COLOR_TITLE_BACKGROUND_ACTIVE, *cfg.COLOUR_TITLE_BACKGROUND)
        imgui.push_style_color(imgui.COLOR_TITLE_BACKGROUND, *cfg.COLOUR_TITLE_BACKGROUND)
        imgui.push_style_color(imgui.COLOR_TITLE_BACKGROUND_COLLAPSED, *cfg.COLOUR_TITLE_BACKGROUND)
        imgui.push_style_color(imgui.COLOR_BUTTON_HOVERED, *cfg.COLOUR_TITLE_BACKGROUND)
        imgui.push_style_color(imgui.COLOR_BUTTON_ACTIVE, *cfg.COLOUR_TITLE_BACKGROUND)
        imgui.push_style_color(imgui.COLOR_FRAME_BACKGROUND, *cfg.COLOUR_FRAME_BACKGROUND)
        imgui.push_style_color(imgui.COLOR_FRAME_BACKGROUND_HOVERED, *cfg.COLOUR_FRAME_BACKGROUND)
        imgui.push_style_color(imgui.COLOR_FRAME_BACKGROUND_ACTIVE, *cfg.COLOUR_FRAME_ACTIVE)
        imgui.push_style_color(imgui.COLOR_BUTTON_ACTIVE, *cfg.COLOUR_FRAME_ACTIVE)
        imgui.push_style_color(imgui.COLOR_BUTTON, *cfg.COLOUR_FRAME_BACKGROUND)
        imgui.push_style_color(imgui.COLOR_BUTTON_HOVERED, *cfg.COLOUR_FRAME_BACKGROUND)
        imgui.push_style_color(imgui.COLOR_SLIDER_GRAB, *cfg.COLOUR_FRAME_EXTRA_DARK)
        imgui.push_style_color(imgui.COLOR_SLIDER_GRAB_ACTIVE, *cfg.COLOUR_FRAME_DARK)
        imgui.push_style_color(imgui.COLOR_POPUP_BACKGROUND, *cfg.COLOUR_FRAME_BACKGROUND)
        imgui.push_style_color(imgui.COLOR_SCROLLBAR_GRAB, *cfg.COLOUR_FRAME_DARK)
        imgui.push_style_color(imgui.COLOR_SCROLLBAR_GRAB_HOVERED, *cfg.COLOUR_FRAME_DARK)
        imgui.push_style_color(imgui.COLOR_SCROLLBAR_GRAB_ACTIVE, *cfg.COLOUR_FRAME_BACKGROUND)
        imgui.push_style_color(imgui.COLOR_SCROLLBAR_BACKGROUND, *cfg.COLOUR_FRAME_EXTRA_DARK)
        imgui.push_style_color(imgui.COLOR_PLOT_HISTOGRAM, *cfg.COLOUR_TEXT)
        imgui.push_style_color(imgui.COLOR_PLOT_HISTOGRAM_HOVERED, *cfg.COLOUR_TEXT)
        imgui.push_style_color(imgui.COLOR_CHECK_MARK, *cfg.COLOUR_FRAME_EXTRA_DARK)
        imgui.push_style_color(imgui.COLOR_MENUBAR_BACKGROUND, *cfg.COLOUR_MAIN_MENU_BAR)
        imgui.push_style_color(imgui.COLOR_HEADER, *cfg.COLOUR_HEADER)
        imgui.push_style_color(imgui.COLOR_HEADER_HOVERED, *cfg.COLOUR_HEADER_HOVERED)
        imgui.push_style_color(imgui.COLOR_HEADER_ACTIVE, *cfg.COLOUR_HEADER_ACTIVE)
        imgui.push_style_color(imgui.COLOR_DRAG_DROP_TARGET, *cfg.COLOUR_DROP_TARGET)
        imgui.push_style_var(imgui.STYLE_WINDOW_ROUNDING, cfg.WINDOW_ROUNDING)

        self._shortcuts()
        self._gui_main()

        imgui.pop_style_color(28)
        imgui.pop_style_var(1)
        imgui.render()
        self.imgui_implementation.render(imgui.get_draw_data())

        if cfg.active_node is not None:
            ImageViewer.MARKER_COLOUR = cfg.active_node.colour
            self.node_blocking_frame_window = cfg.active_node.DISABLE_FRAME_INFO_WINDOW
            if self.previous_active_node is not cfg.active_node or cfg.active_node.any_change or self.new_image_requested:
                if cfg.active_node.NODE_RETURNS_IMAGE:
                    an = cfg.active_node
                    self.current_dataset = an.get_source_load_data_node(an).dataset
                    #if self.current_dataset.initialized: removed this if clause 221017 - final else was: self.show-iamge = False
                    self.current_dataset.current_frame = np.clip(self.current_dataset.current_frame, 0, self.current_dataset.n_frames - 1)
                    cfg.active_node.FRAME_REQUESTED_BY_IMAGE_VIEWER = True
                    new_image = cfg.active_node.get_image(self.current_dataset.current_frame)
                    cfg.active_node.FRAME_REQUESTED_BY_IMAGE_VIEWER = False
                    if cfg.active_node.OVERRIDE_AUTOCONTRAST:
                        print("Overriding contrast settings")
                        self.autocontrast = [False, False, False]
                        lims = cfg.active_node.OVERRIDE_AUTOCONTRAST_LIMS
                        self.contrast_min = [lims[0], lims[0], lims[0]]
                        self.contrast_max = [lims[1], lims[1], lims[1]]
                    if new_image is not None:
                        self.show_image = True
                        self.set_image(new_image)
                    else:
                        self.show_image = False
            # set node-specific LUT
            if cfg.active_node.lut != "auto":
                self.node_specific_lut_in_use = True
                self.set_lut(settings.lut_names.index(cfg.active_node.lut))
            else:
                self.node_specific_lut_in_use = False
                self.set_lut(self.base_lut)
            self.previous_active_node = cfg.active_node

    def end_frame(self):
        self.window.end_frame()

    def close(self):
        self.imgui_implementation.shutdown()

    def _gui_main(self):
        if self.show_frame_select_window and self.current_dataset.initialized and not self.mode == "RGB" and not self.node_blocking_frame_window:
            self._frame_info_window()

        # Context menu
        if not self.context_menu_open:
            if self.window.get_mouse_event(glfw.MOUSE_BUTTON_RIGHT, glfw.PRESS, 0):
                self.context_menu_position = self.window.cursor_pos
                self.context_menu_open = True
                self.context_menu_can_close = False
        else:
            self._context_menu()

        # Brightness/contrast window
        if self.contrast_window_open:
            imgui.set_next_window_size(ImageViewer.CONTRAST_WINDOW_SIZE[0], ImageViewer.CONTRAST_WINDOW_SIZE[1])
            imgui.set_next_window_position(self.contrast_window_position[0], self.contrast_window_position[1], condition = imgui.APPEARING)
            _, self.contrast_window_open = imgui.begin("Adjust contrast", closable = True, flags = imgui.WINDOW_NO_COLLAPSE | imgui.WINDOW_NO_RESIZE)
            self.contrast_window_position = imgui.get_window_position()
            available_width = imgui.get_content_region_available_width()
            _rgb_button_spacing = 4

            if self.mode == "RGB":
                if imgui.button("R", width=(available_width - 2 * _rgb_button_spacing) / 3, height=30):
                    self.contrast_window_channel = 0
                imgui.same_line(spacing = _rgb_button_spacing)
                if imgui.button("G", width=(available_width - 2 * _rgb_button_spacing) / 3, height=30):
                    self.contrast_window_channel = 1
                imgui.same_line(spacing=_rgb_button_spacing)
                if imgui.button("B", width=(available_width - 2 * _rgb_button_spacing) / 3, height=30):
                    self.contrast_window_channel = 2

            if self.mode == "R":
                self.contrast_window_channel = 0
                imgui.push_style_color(imgui.COLOR_PLOT_HISTOGRAM, 0.0, 0.0, 0.0, 1.0)
                imgui.push_style_color(imgui.COLOR_PLOT_HISTOGRAM_HOVERED, 0.0, 0.0, 0.0, 1.0)
                imgui.push_style_color(imgui.COLOR_SLIDER_GRAB, 0.0, 0.0, 0.0, 1.0)
                imgui.push_style_color(imgui.COLOR_SLIDER_GRAB_ACTIVE, 0.0, 0.0, 0.0, 1.0)
            else:
                _clr = (float(self.contrast_window_channel == 0) + 0.2 * 0.8, float(self.contrast_window_channel == 1) + 0.2 * 0.8, float(self.contrast_window_channel == 2) + 0.2 * 0.8)
                imgui.push_style_color(imgui.COLOR_PLOT_HISTOGRAM, *_clr)
                imgui.push_style_color(imgui.COLOR_PLOT_HISTOGRAM_HOVERED, *_clr)
                imgui.push_style_color(imgui.COLOR_SLIDER_GRAB, *_clr)
                imgui.push_style_color(imgui.COLOR_SLIDER_GRAB_ACTIVE, *_clr)


            imgui.plot_histogram("##contrast_hist", self.hist_counts[self.contrast_window_channel], graph_size = (available_width, 100))
            imgui.push_item_width(available_width)
            _min_changed, self.contrast_min[self.contrast_window_channel] = imgui.slider_float("##min_contrast", self.contrast_min[self.contrast_window_channel], self.image_amin[self.contrast_window_channel], self.image_amax[self.contrast_window_channel] + 1, format="min: %1.0f")
            _max_changed, self.contrast_max[self.contrast_window_channel] = imgui.slider_float("##max_contrast", self.contrast_max[self.contrast_window_channel], self.image_amin[self.contrast_window_channel], self.image_amax[self.contrast_window_channel] + 1, format="max: %1.0f")
            if _min_changed or _max_changed:
                self.autocontrast[self.contrast_window_channel] = False
            imgui.pop_item_width()
            if imgui.button("Auto once", width = 80, height = ImageViewer.FRAME_SELECT_BUTTON_SIZE[1]):
                self._compute_auto_contrast(channel = self.contrast_window_channel)
            imgui.same_line(spacing = ImageViewer.FRAME_SELECT_BUTTON_SPACING)
            _always_auto_changed, self.autocontrast[self.contrast_window_channel] = imgui.checkbox("always auto", self.autocontrast[self.contrast_window_channel])
            if _always_auto_changed and self.autocontrast[self.contrast_window_channel]:
                self._compute_auto_contrast()
            if imgui.is_window_focused():
                if self.window.get_key_event(glfw.KEY_SPACE, glfw.PRESS):
                    self._compute_auto_contrast()
            imgui.pop_style_color(4)
            imgui.end()

        # No image provided by active node
        if self.show_image is False:
            imgui.set_next_window_position(self.window.width // 2 - ImageViewer.NO_IMAGE_NOTIFICATION_SIZE[0] // 2, self.window.height // 2 - ImageViewer.NO_IMAGE_NOTIFICATION_SIZE[1] // 2)
            imgui.set_next_window_size(*ImageViewer.NO_IMAGE_NOTIFICATION_SIZE)
            imgui.push_style_color(imgui.COLOR_WINDOW_BACKGROUND, *ImageViewer.COLOUR_CLEAR)
            imgui.push_style_color(imgui.COLOR_BORDER, *ImageViewer.COLOUR_CLEAR)
            imgui.push_style_color(imgui.COLOR_TEXT, *(0.0, 0.0, 0.0, 1.0))
            imgui.begin("##", False, imgui.WINDOW_NO_TITLE_BAR | imgui.WINDOW_NO_RESIZE | imgui.WINDOW_NO_TITLE_BAR | imgui.WINDOW_NO_COLLAPSE | imgui.WINDOW_NO_BACKGROUND)
            imgui.text("   Active node does not output an image.")
            imgui.end()
            imgui.pop_style_color(3)

    def _render(self):
        if self.show_image:
            self.va.bind()
            self.shader.bind()
            self.texture.bind(0)
            self.lut_texture.bind(1)
            self.shader.uniformmat4("cameraMatrix", self.camera.view_projection_matrix)
            self.shader.uniform3f("contrast_min", self.contrast_min)
            self.shader.uniform3f("contrast_max", self.contrast_max)
            self.shader.uniform3f("translation", [0.0, 0.0, 0.0])
            self.shader.uniform1i("use_lut", 1 if self.mode == "R" else 0)
            glDrawElements(GL_TRIANGLES, self.va.indexBuffer.getCount(), GL_UNSIGNED_SHORT, None)
            self.shader.unbind()
            self.va.unbind()
            glActiveTexture(GL_TEXTURE0)
            self.marker.render_start(self.roi_shader, self.camera, ImageViewer.MARKER_COLOUR)
            for coordinate in self.image.maxima:
                translation = [coordinate[1], coordinate[0]]
                self.marker.render(self.roi_shader, translation)
            self.marker.render_end(self.roi_shader)

    def _frame_info_window(self):

        imgui.push_style_var(imgui.STYLE_WINDOW_ROUNDING, 0.0)
        imgui.set_next_window_position(0, self.window.height - ImageViewer.INFO_BAR_HEIGHT + (cfg.window_height - self.window.height), imgui.ALWAYS)
        imgui.set_next_window_size(self.window.width, ImageViewer.INFO_BAR_HEIGHT)
        ## Info & control panel
        imgui.begin("##frameselectwindow", flags=imgui.WINDOW_NO_TITLE_BAR | imgui.WINDOW_NO_RESIZE | imgui.WINDOW_NO_MOVE)

        # Frame slider and buttons
        available_width = imgui.get_content_region_available_width()
        _frame_idx_changed = False
        if imgui.button("<", width=ImageViewer.FRAME_SELECT_BUTTON_SIZE[0], height=ImageViewer.FRAME_SELECT_BUTTON_SIZE[1]):
            self.current_dataset.current_frame -= 1
            _frame_idx_changed = True
        imgui.same_line(spacing=ImageViewer.FRAME_SELECT_BUTTON_SPACING)
        imgui.push_item_width(
            available_width - 2 * ImageViewer.FRAME_SELECT_BUTTON_SIZE[0] - 2 * ImageViewer.FRAME_SELECT_BUTTON_SPACING)

        _slider_changed, self.current_dataset.current_frame = imgui.slider_int("##frame_select",
                                                                               self.current_dataset.current_frame, 0,
                                                                               self.current_dataset.n_frames - 1,
                                                                               f"Frame {self.current_dataset.current_frame}/{self.current_dataset.n_frames}")

        _frame_idx_changed = _slider_changed or _frame_idx_changed
        imgui.same_line(spacing=ImageViewer.FRAME_SELECT_BUTTON_SPACING)
        if imgui.button(">", width=ImageViewer.FRAME_SELECT_BUTTON_SIZE[0], height=ImageViewer.FRAME_SELECT_BUTTON_SIZE[1]):
            self.current_dataset.current_frame += 1
            _frame_idx_changed = True
        if not self.window.get_key(glfw.KEY_LEFT_SHIFT):
            if self.window.scroll_delta[1] != 0:
                _frame_idx_changed = True
                self.current_dataset.current_frame -= int(self.window.scroll_delta[1])
        if self.window.get_key_event(glfw.KEY_LEFT, glfw.PRESS) or self.window.get_key_event(glfw.KEY_LEFT, glfw.REPEAT):
            _frame_idx_changed = True
            self.current_dataset.current_frame -= 1
        if self.window.get_key_event(glfw.KEY_RIGHT, glfw.PRESS) or self.window.get_key_event(glfw.KEY_RIGHT, glfw.REPEAT):
            _frame_idx_changed = True
            self.current_dataset.current_frame += 1
        if _frame_idx_changed:
            self.current_dataset.current_frame = np.clip(self.current_dataset.current_frame, 0,
                                                         self.current_dataset.n_frames - 1)
            self.new_image_requested = True

        imgui.pop_item_width()

        # Image info
        imgui.separator()
        imgui.text(self.frame_info)
        imgui.pop_style_var(1)
        imgui.end()


    def _edit_and_render_roi(self):
        if not self.show_image:
            return None
        if cfg.active_node is None:
            return None
        if not cfg.active_node.use_roi:
            return None
        # if past the above, active node has and used a roi.
        if cfg.active_node.roi == [0, 0, 0, 0]:
            cfg.active_node.roi = [self.image_width // 4, self.image_height // 4, self.image_width * 3 // 4, self.image_height * 3 // 4]
        else:
            self.roi.set_box(cfg.active_node.roi)
            self.roi.colour = cfg.active_node.colour
            self.roi.render(self.roi_shader, self.camera)
            # edit roi
            if not imgui.get_io().want_capture_mouse and not self.window.window_gained_focus:
                if self.window.get_mouse_event(glfw.MOUSE_BUTTON_LEFT, glfw.PRESS):
                    # get world position of mouse click.
                    px_coords = self.get_cursor_image_coordinates()
                    if self.roi.is_in_roi(px_coords):
                        self.moving_roi = True
                    else:
                        self.drawing_roi = True
                        self.roi.set_box([*px_coords, *px_coords])

                if self.moving_roi:
                    if self.window.get_mouse_event(glfw.MOUSE_BUTTON_LEFT, glfw.RELEASE):
                        self.moving_roi = False
                        self.roi.limit(self.image_width, self.image_height)
                        cfg.active_node.any_change = True
                    else:
                        self.roi.translate(self.cursor_delta_as_world_delta())

                elif self.drawing_roi:
                    px_coords = self.get_cursor_image_coordinates()
                    if self.window.get_mouse_event(glfw.MOUSE_BUTTON_LEFT, glfw.RELEASE):
                        self.drawing_roi = False
                        self.roi.correct_order()
                        self.roi.limit(self.image_width, self.image_height)
                        cfg.active_node.any_change = True
                    else:
                        new_box = [self.roi.box[0], self.roi.box[1], px_coords[0], px_coords[1]]
                        if cfg.active_node.ROI_MUST_BE_SQUARE:
                            box_width = np.abs(px_coords[0] - self.roi.box[0])
                            box_height = np.abs(px_coords[1] - self.roi.box[1])
                            box_size = min([box_width, box_height])
                            new_box = [self.roi.box[0], self.roi.box[1], self.roi.box[0] + box_size, self.roi.box[1] + box_size]
                        self.roi.set_box(new_box)
            box = self.roi.box
            if not self.drawing_roi:
                if box[0] == box[2] and box[1] != box[3] or box[1] == box[3] and box[0] != box[2]:
                    self.roi.set_box([self.image_width // 4, self.image_height // 4, self.image_width * 3 // 4, self.image_height * 3 // 4])
            cfg.active_node.roi = self.roi.box

    def _camera_control(self):
        if not imgui.get_io().want_capture_mouse:
            if self.window.get_mouse_button(glfw.MOUSE_BUTTON_MIDDLE):
                self.camera.position[0] += self.window.cursor_delta[0] * ImageViewer.CAMERA_PAN_SPEED
                self.camera.position[1] -= self.window.cursor_delta[1] * ImageViewer.CAMERA_PAN_SPEED
            if self.window.get_key(glfw.KEY_LEFT_SHIFT):
                if self.window.scroll_delta[1] != 0:
                    camera_updated_zoom = self.camera.zoom * (1.0 + self.window.scroll_delta[1] * ImageViewer.CAMERA_ZOOM_STEP)
                    if camera_updated_zoom < ImageViewer.CAMERA_ZOOM_MAX:
                        self.camera.zoom *= (1.0 + self.window.scroll_delta[1] * ImageViewer.CAMERA_ZOOM_STEP)
                        self.camera.position[0] *= (1.0 + self.window.scroll_delta[1] * ImageViewer.CAMERA_ZOOM_STEP)
                        self.camera.position[1] *= (1.0 + self.window.scroll_delta[1] * ImageViewer.CAMERA_ZOOM_STEP)

    def _shortcuts(self):
        if self.window.get_key_event(glfw.KEY_C, glfw.PRESS, glfw.MOD_SHIFT | glfw.MOD_CONTROL):
            self.contrast_window_open = True
        if self.window.get_key_event(glfw.KEY_W, glfw.PRESS, glfw.MOD_CONTROL):
            self.contrast_window_open = False
        if self.contrast_window_open:
            if self.window.get_key_event(glfw.KEY_SPACE, glfw.PRESS, 0):
                self._compute_auto_contrast(0)
                self._compute_auto_contrast(1)
                self._compute_auto_contrast(2)
            if self.window.get_key_event(glfw.KEY_SPACE, glfw.PRESS, glfw.MOD_SHIFT | glfw.MOD_CONTROL):
                self.autocontrast = [True, True, True]
        if self.window.get_key_event(glfw.KEY_S, glfw.PRESS, glfw.MOD_CONTROL):
            if self.image_pxd is not None:
                self.save_current_image()
        if self.window.get_key_event(glfw.KEY_DELETE, glfw.PRESS, 0) or self.window.get_key_event(glfw.KEY_DELETE, glfw.REPEAT, 0):
            self.current_dataset.delete_by_index(self.current_dataset.current_frame)
            self.new_image_requested = True
        if self.window.get_key_event(glfw.KEY_TAB, glfw.PRESS):
            cfg.active_editor = 1 - cfg.active_editor

    def _context_menu(self):
        imgui.set_next_window_position(self.context_menu_position[0] - 3, self.context_menu_position[1] - 3)
        imgui.begin("##ivcontextmenu", flags=imgui.WINDOW_NO_MOVE | imgui.WINDOW_NO_TITLE_BAR | imgui.WINDOW_NO_RESIZE | imgui.WINDOW_ALWAYS_AUTO_RESIZE)
        # Close context menu when it is not hovered.
        context_menu_hovered = imgui.is_window_hovered(flags=imgui.HOVERED_ALLOW_WHEN_BLOCKED_BY_POPUP | imgui.HOVERED_CHILD_WINDOWS)
        if context_menu_hovered:
            self.context_menu_can_close = True
        if not context_menu_hovered and self.context_menu_can_close:
            self.context_menu_can_close = False
            self.context_menu_open = False

        _open, _ = imgui.menu_item("Brightness / contrast")
        if _open:
            self.contrast_window_open = True
            self.contrast_window_position = self.context_menu_position
            self.context_menu_open = False
        _reset_view, _ = imgui.menu_item("Reset view")
        if _reset_view:
            self.center_camera()
            self.context_menu_open = False
        _add_to_correlation_editor, _ = imgui.menu_item("Add to Correlation Editor")
        if _add_to_correlation_editor:
            self.image._ce_lut = self.current_lut + 1  # note: idx + 1 as ImageViewer has an extra 'auto' lut option at index 0.
            self.image._ce_clims = [self.contrast_min[0], self.contrast_max[0], self.contrast_min[1], self.contrast_max[1], self.contrast_min[2], self.contrast_max[2]]
            cfg.correlation_editor.add_frame(self.image)
            self.context_menu_open = False
        _export, _ = imgui.menu_item("Save as .tiff")
        if _export:
            self.save_current_image()
            self.context_menu_open = False
        if imgui.begin_menu("Change LUT"):
            for key in settings.lut_names:
                key_clicked, _ = imgui.menu_item(key)
                if key_clicked:
                    self.set_lut(settings.lut_names.index(key))
            imgui.end_menu()
        imgui.end()

    def set_image( self, image):
        self.image_requires_update = True
        if type(image) == np.ndarray:
            self.image = Frame("virtual_path")
            self.image.data = image
        else:
            self.image = image
        self.frame_info = str(self.image)

    def update_image(self):
        self.image_pxd = self.image.load()
        width = np.shape(self.image_pxd)[1]
        height = np.shape(self.image_pxd)[0]
        if len(self.image_pxd.shape) == 2:
            self.image_pxd = np.repeat(self.image_pxd[:, :, np.newaxis], 3, axis=2)
            self.set_mode("R")
        else:
            self.set_mode("RGB")
        self.image_amax = np.amax(self.image_pxd, axis = (0, 1))
        self.image_amin = np.amin(self.image_pxd, axis = (0, 1))
        if self.autocontrast[0]:
            self._compute_auto_contrast(0)
        if self.autocontrast[1]:
            self._compute_auto_contrast(1)
        if self.autocontrast[1]:
            self._compute_auto_contrast(2)
        if self.image_width != width or self.image_height != height:
            self.image_size_changed = True
            self.image_width = width
            self.image_height = height

            vertex_attributes = [0.0, height, 1.0, 0.0, 1.0,
                                 0.0, 0.0, 1.0, 0.0, 0.0,
                                 width, 0.0, 1.0, 1.0, 0.0,
                                 width, height, 1.0, 1.0, 1.0]
            indices = [0, 1, 2, 2, 0, 3]
            self.va.update(VertexBuffer(vertex_attributes), IndexBuffer(indices))
        self.texture.update(self.image_pxd.astype(np.float))

        self.hist_counts[0], self.hist_bins[0] = np.histogram(self.image_pxd[:, :, 0], bins=ImageViewer.HISTOGRAM_BINS)
        self.hist_counts[0] = self.hist_counts[0].astype('float32')
        self.hist_counts[0] = np.delete(self.hist_counts[0], 0)
        self.hist_bins[0] = np.delete(self.hist_bins[0], 0)
        self.hist_counts[0] = np.log(self.hist_counts[0] + 1)

        if self.mode == "RGB":
            self.hist_counts[1], self.hist_bins[1] = np.histogram(self.image_pxd[:, :, 1], bins=ImageViewer.HISTOGRAM_BINS)
            self.hist_counts[1] = self.hist_counts[1].astype('float32')
            self.hist_counts[1] = np.delete(self.hist_counts[1], 0)
            self.hist_bins[1] = np.delete(self.hist_bins[1], 0)

            self.hist_counts[2], self.hist_bins[2] = np.histogram(self.image_pxd[:, :, 2], bins=ImageViewer.HISTOGRAM_BINS)
            self.hist_counts[2] = self.hist_counts[2].astype('float32')
            self.hist_counts[2] = np.delete(self.hist_counts[2], 0)
            self.hist_bins[2] = np.delete(self.hist_bins[2], 0)
            self.hist_counts[1] = np.log(self.hist_counts[1] + 1)
            self.hist_counts[2] = np.log(self.hist_counts[2] + 1)
    def _compute_auto_contrast(self, channel=None):
        img_subsample = self.image_pxd[::settings.autocontrast_subsample, ::settings.autocontrast_subsample, :]
        n = img_subsample.shape[0] * img_subsample.shape[1]
        for i in range(3):
            if self.autocontrast[i] or channel == i:
                img_sorted = np.sort(img_subsample[:, :, i].flatten())
                self.contrast_min[i] = img_sorted[int(settings.autocontrast_saturation / 100.0 * n)]
                self.contrast_max[i] = img_sorted[int((1.0 - settings.autocontrast_saturation / 100.0) * n)]

    def center_camera(self):
        self.camera.zoom = 1
        self.camera.position = [-self.image_width / 2, -self.image_height / 2, 0.0]

    def set_lut(self, lut_index):
        if not self.node_specific_lut_in_use:
            self.base_lut = lut_index
        if lut_index != self.current_lut:
            self.current_lut = lut_index
            self.lut_array = np.asarray(settings.luts[settings.lut_names[self.current_lut]])
            if self.lut_array.shape[1] == 3:
                self.lut_array = np.reshape(self.lut_array, (self.lut_array.shape[0], 1, self.lut_array.shape[1]))  ## was lut_array = -> now self.lut_array
            self.lut_texture.update(self.lut_array)

    def get_cursor_image_coordinates(self):
        c_pos = self.window.cursor_pos  # cursor position
        c_pos_ndc = [(c_pos[0] / self.window.width) * 2.0 - 1.0, -((c_pos[1] / self.window.height) * 2.0 - 1.0)]
        world_pos = np.matmul(self.camera.view_matrix.I, self.camera.projection_matrix.I) \
                    * np.matrix([[c_pos_ndc[0]], [c_pos_ndc[1]], [1], [1]])
        image_coordinates = [int(world_pos[0]), int(world_pos[1])]
        return image_coordinates

    def cursor_delta_as_world_delta(self):
        return [self.window.cursor_delta[0] / self.camera.zoom, -self.window.cursor_delta[1] / self.camera.zoom]

    def save_current_image(self):
        try:
            filename = filedialog.asksaveasfilename()
            if filename != '':
                if self.mode == "R":
                    save_tiff(self.image_pxd[:, :, 0], filename, self.image.pixel_size)
                else:
                    save_tiff(self.image_pxd, filename, self.image.pixel_size)
        except Exception as e:
            cfg.set_error(e, f"Error saving file via ImageViewer.save_current_image():\n{str(e)}")


class Camera:
    def __init__(self):
        self.position = np.asarray([0.0, 0.0, 0.0])
        self.zoom = 1.0
        self.view_matrix = np.identity(4)
        self.projection_matrix = np.identity(4)
        self.view_projection_matrix = np.identity(4)
        self.set_projection_matrix(cfg.iv_window_width, cfg.iv_window_height)

    def set_projection_matrix(self, window_width, window_height):
        self.projection_matrix = np.matrix([
            [2 / window_width, 0, 0, 0],
            [0, 2 / window_height, 0, 0],
            [0, 0, -2 / 100, 0],
            [0, 0, 0, 1],
        ])

    def on_update(self):
        self.view_matrix = np.matrix([
            [self.zoom, 0.0, 0.0, self.position[0]],
            [0.0, self.zoom, 0.0, self.position[1]],
            [0.0, 0.0, self.zoom, self.position[2]],
            [0.0, 0.0, 0.0, 1.0],
        ])
        self.view_projection_matrix = np.matmul(self.projection_matrix, self.view_matrix)






from setuptools import setup, find_packages

setup(
    name='scNodes',
    version='1.0.56',
    packages=find_packages(),
    url='https://github.com/bionanopatterning/scNodes',
    license='GPL v3',
    author='mgflast',
    author_email='m.g.f.last@lumc.nl',
    description='Correlating super-resolution fluorescence and transmission electron cryo-microscopy',
    package_data={'': ['*.png', '*.glsl', '*.whl', '*.tar.gz', '*.pdf']},
    include_package_data=False,  # weirdly, the above filetypes are only included when this parameter is set to False.
    install_requires=[
        "colorcet>=3.0.0",
        "dill>=0.3.5.1",
        "glfw>=2.5.5",
        "joblib>=1.1.0",
        "matplotlib>=3.5.3",
        "mrcfile>=1.4.3",
        "numpy>=1.3.0",  ## todo check - was 1.23.2
        "opencv-python>=4.6.0.66",
        "pandas>=1.3.5",  ## todo check - was 1.4.4
        "Pillow>=9.2.0",
        "psutil>=5.9.2",
        "PyOpenGl>=3.1.6",
        "PyWavelets>=1.3.0",
        "pyperclip>=1.8.2",
        "pystackreg>=0.2.6.post1",
        "pyimgui-wheels",
        "scikit-image>=0.19.3",
        "tifffile>=2021.11.2" ## todo check - was 2022.8.12
    ]
)



try:
    from . import finterp as f
    from . import filtering
    from . import switches as s
except:
    import finterp as f
    import filtering
    import switches as s

import numpy as np

if s.SPHINX_SWITCH is False:
    import tifffile as tiff

import scipy.special
import sys
import os
import collections
import math
import itertools


def sorted_k_partitions(seq, k):
    """
    Returns a list of all unique k-partitions of `seq`.
    Each partition is a list of parts, and each part is a tuple.
    """
    n = len(seq)
    groups = []

    def generate_partitions(i):
        if i >= n:
            yield list(map(tuple, groups))
        else:
            if n - i > k - len(groups):
                for group in groups:
                    group.append(seq[i])
                    yield from generate_partitions(i + 1)
                    group.pop()
            if len(groups) < k:
                groups.append([seq[i]])
                yield from generate_partitions(i + 1)
                groups.pop()

    result = generate_partitions(0)

    # Sort the parts in each partition in shortlex order
    result = [sorted(ps, key=lambda p: (len(p), p)) for ps in result]
    # Sort partitions by the length of each part, then lexicographically.
    result = sorted(result, key=lambda ps: (*map(len, ps), ps))
    # delete partitions with only 1 element
    result = [p for p in result if len(p[0]) > 1]

    return result


def average_image(filepath, filename, frames=[]):
    """
    Get the average image for a video file (tiff stack), either for the
    whole video or for user defined frames.

    Parameters
    ----------
    filepath : str
        Path to the tiff file.
    filename : str
        Name of the tiff file.
    frames : list of int, optional
        Start and end frame number if not the whole video is used.

    Returns
    -------
    mean_im : ndarray
        The average image.
    """
    imstack = tiff.TiffFile(filepath + '/' + filename)
    xdim, ydim = np.shape(imstack.pages[0])
    mean_im = np.zeros((xdim, ydim))

    if not frames:
        frames = [0, len(imstack.pages)]
    for frame_num in range(frames[0], frames[1]):
        im = tiff.imread(filepath + '/' + filename, key=frame_num)
        mean_im = mean_im + im
        sys.stdout.write('\r')
        sys.stdout.write("frame " + str(frame_num) + '/' + str(frames[1]))
        sys.stdout.flush()
    mean_im = mean_im / (frames[1] - frames[0])
    imstack.close()

    return mean_im


def average_image_with_finterp(filepath, filename, interp_num):
    """
    Get the average image with fourier interpolation for a video file.

    Parameters
    ----------
    filepath : str
        Path to the tiff file.
    filename : str
        Name of the tiff file.
    interp_num : int
        Interpolation factor.

    Returns
    -------
    mean_im : ndarray
        The average image after fourier interpolation. Interpolated
    images can be further used for SOFI processing.
    """
    original_mean_im = average_image(filepath, filename)
    finterp_mean_im = f.fourier_interp_array(original_mean_im, [interp_num])

    return finterp_mean_im[0]


def calc_moment_im(filepath, filename, order, frames=[], mean_im=None):
    """
    Get one moment-reconstructed image of a defined order for a video file.

    Parameters
    ----------
    filepath : str
        Path to the tiff file.
    filename : str
        Name of the tiff file.
    order : int
        The order number of moment-reconstructed image.
    frames : list of int
        The start and end frame number.
    mean_im : ndarray
        Average image of the tiff stack.

    Returns
    -------
    moment_im : ndarray
        The moments-reconstructed image.
    """
    if mean_im is None:
        mean_im = average_image(filepath, filename, frames)
    imstack = tiff.TiffFile(filepath + '/' + filename)
    xdim, ydim = np.shape(imstack.pages[0])
    moment_im = np.zeros((xdim, ydim))
    print('Calculating the %s-order moment ...' %
          order)
    if frames:
        for frame_num in range(frames[0], frames[1]):
            im = tiff.imread(filepath + '/' + filename, key=frame_num)
            moment_im = moment_im + (im - mean_im) ** order
            sys.stdout.write('\r')
            sys.stdout.write("[{:{}}] {:.1f}%".format(
                "=" * int(30 / (frames[1] - frames[0]) * (frame_num - frames[0] + 1)), 29,
                (100 / (frames[1] - frames[0]) * (frame_num - frames[0] + 1))))
            sys.stdout.flush()
        moment_im = moment_im / (frames[1] - frames[0])
    else:
        mvlength = len(imstack.pages)
        for frame_num in range(mvlength):
            im = tiff.imread(filepath + '/' + filename, key=frame_num)
            moment_im = moment_im + (im - mean_im) ** order
            sys.stdout.write('\r')
            sys.stdout.write("[{:{}}] {:.1f}%".format(
                "=" * int(30 / mvlength * (frame_num + 1)), 29,
                (100 / mvlength * (frame_num + 1))))
            sys.stdout.flush()
        moment_im = moment_im / mvlength
    imstack.close()
    print('\n')
    return moment_im


def moment_im_with_finterp(filepath, filename, order, interp_num,
                           frames=[], mean_im=None):
    """
    Get one moment-reconstructed image of a defined order for a video file.

    Parameters
    ----------
    filepath : str
        Path to the tiff file.
    filename : str
        Name of the tiff file.
    order : int
        The order number of moment-reconstructed image.
    interp_num : int
        The interpolation factor.
    mvlength : int
        The length of video for the reconstruction.
    mean_im : ndarray
        Average image of the tiff stack.

    Returns
    -------
    moment_im : ndarray
        The moments-reconstructed image.
    """
    if mean_im is None:
        mean_im = average_image_with_finterp(filepath, filename, interp_num)

    imstack = tiff.TiffFile(filepath + '/' + filename)
    xdim, ydim = np.shape(imstack.pages[0])
    moment_im = np.zeros(((xdim - 1) * interp_num + 1, (ydim - 1) * interp_num + 1))
    if frames:
        for frame_num in range(frames[0], frames[1]):
            im = tiff.imread(filepath + '/' + filename, key=frame_num)
            interp_im = f.fourier_interp_array(im, [interp_num])[0]
            moment_im = moment_im + (interp_im - mean_im) ** order
            sys.stdout.write('\r')
            sys.stdout.write("[{:{}}] {:.1f}%".format(
                "=" * int(30 / (frames[1] - frames[0]) * (frame_num - frames[0] + 1)), 29,
                (100 / (frames[1] - frames[0]) * (frame_num - frames[0] + 1))))
            sys.stdout.flush()
        moment_im = np.int64(moment_im / (frames[1] - frames[0]))
    else:
        mvlength = len(imstack.pages)
        for frame_num in range(mvlength):
            im = tiff.imread(filepath + '/' + filename, key=frame_num)
            interp_im = f.fourier_interp_array(im, [interp_num])[0]
            moment_im = moment_im + (interp_im - mean_im) ** order
            sys.stdout.write('\r')
            sys.stdout.write("[{:{}}] {:.1f}%".format(
                "=" * int(30 / mvlength * (frame_num + 1)), 29,
                (100 / mvlength * (frame_num + 1))))
            sys.stdout.flush()
        moment_im = np.int64(moment_im / mvlength)
    imstack.close()
    return moment_im


def calc_moments(filepath, filename, highest_order,
                 frames=None, m_set=None, mean_im=None):
    """
    Get all moment-reconstructed images to the user-defined highest order for
    a video file(tiff stack).

    Parameters
    ----------
    filepath : str
        Path to the tiff file.
    filename : str
        Name of the tiff file.
    highest_order : int
        The highest order number of moment-reconstructed images.
    frames : list of int
        The start and end frame number.
    m_set : dict
        order number (int) -> image (ndarray)
        A dictionary of previously calcualted moment-reconstructed images.
    mean_im : ndarray
        Average image of the tiff stack.

    Returns
    -------
    m_set : dict
        order number (int) -> image (ndarray)
        A dictionary of calcualted moment-reconstructed images.
    """
    if m_set is None:
        m_set = {}
    if frames is None:
        frames = []
    if m_set:
        current_order = max(m_set.keys())
    else:
        current_order = 0

    # calculate average image (mean)
    if mean_im is None:
        mean_im = average_image(filepath, filename, frames)

    # load in the full image stack
    imstack = tiff.TiffFile(filepath + '/' + filename)

    # get the dimension of each frame
    xdim, ydim = np.shape(imstack.pages[0])

    def ordinal(n):
        return "%d%s" % (
            n, "tsnrhtdd"[(n // 10 % 10 != 1) * (n % 10 < 4) * n % 10::4])

    order_lst = [ordinal(n + 1) for n in range(highest_order)]

    if highest_order > current_order:
        for order in range(current_order, highest_order):
            print('Calculating the %s-order moment...' %
                  order_lst[order])
            m_set[order + 1] = np.zeros((xdim, ydim))
            if frames:  # if the start and end frame number is specified.
                for frame_num in range(frames[0], frames[1]):
                    im = tiff.imread(filepath + '/' + filename, key=frame_num)
                    m_set[order + 1] = m_set[order + 1] + \
                                       np.power(im - mean_im, order + 1)
                    sys.stdout.write('\r')
                    sys.stdout.write("[{:{}}] {:.1f}%".format(
                        "=" * int(30 / (frames[1] - frames[0]) * (frame_num - frames[0] + 1)),
                        29, (100 / (frames[1] - frames[0]) * (frame_num - frames[0] + 1))))
                    sys.stdout.flush()
                m_set[order + 1] = np.int64(
                    m_set[order + 1] / (frames[1] - frames[0]))  # dvide by the total number of frames
                print('\n')
            else:
                mvlength = len(imstack.pages)
                for frame_num in range(mvlength):
                    im = tiff.imread(filepath + '/' + filename, key=frame_num)
                    m_set[order + 1] = m_set[order + 1] + \
                                       np.power(im - mean_im, order + 1)
                    sys.stdout.write('\r')
                    sys.stdout.write("[{:{}}] {:.1f}%".format(
                        "=" * int(30 / mvlength * (frame_num + 1)), 29,
                        (100 / mvlength * (frame_num + 1))))
                    sys.stdout.flush()
                m_set[order + 1] = np.int64(m_set[order + 1] / mvlength)
                print('\n')
    imstack.close()
    return m_set


def calc_cumulants_from_moments(moment_set):
    """
    Calculate cumulant images from moment images using the recursive relation.

    Parameters
    ----------
    moment_set : dict
        order number (int) -> image (ndarray)
        A dictionary of calcualted moment- images.

    Returns
    -------
    k_set : dict
        order number (int) -> image (ndarray)
        A dictionary of calcualted cumulant-reconstructed images.
    """
    if moment_set == {}:
        raise Exception("'moment_set' is empty.")

    k_set = {}
    highest_order = max(moment_set.keys())
    for order in range(1, highest_order + 1):
        k_set[order] = moment_set[order] - \
                       np.sum(np.array([scipy.special.comb(order - 1, i) * k_set[order - i] *
                                        moment_set[i] for i in range(1, order)]), axis=0)

    return k_set


def calc_block_moments(filepath, filename, highest_order, frames=[]):
    """
    Get moment-reconstructed images for user-defined frames (block) of
    a video file(tiff stack).

    Parameters
    ----------
    filepath : str
        Path to the tiff file.
    filename : str
        Name of the tiff file.
    highest_order : int
        The highest order number of moment-reconstructed images.
    frames : list of int
        Start and end frame number.

    Returns
    -------
    m_set : dict
        order number (int) -> image (ndarray)
        A dictionary of calcualted moment-reconstructed images.

    Notes
    -----
    Similar to 'calc_moments'. Here we omit previously calculated m_set
    and mean_im as inputs since a block usually has much fewer number of
    frames and takes shorter calculation time.
    """
    mean_im = average_image(filepath, filename, frames)
    imstack = tiff.TiffFile(filepath + '/' + filename)
    if not frames:
        mvlength = len(imstack.pages)
        frames = [0, mvlength]
    xdim, ydim = np.shape(imstack.pages[0])
    block_length = frames[1] - frames[0]
    m_set = {}

    for order in range(highest_order):
        m_set[order + 1] = np.zeros((xdim, ydim))
        for frame_num in range(frames[0], frames[1]):
            im = tiff.imread(filepath + '/' + filename, key=frame_num)
            m_set[order + 1] = m_set[order + 1] + \
                               np.power(im - mean_im, order + 1)
        m_set[order + 1] = m_set[order + 1] / block_length
        sys.stdout.write('\r')
        sys.stdout.write("[{:{}}] {:.1f}%".format(
            "=" * int(30 / (highest_order - 1) * order), 29,
            (100 / (highest_order - 1) * order)))
        sys.stdout.flush()
    imstack.close()
    print('\n')
    return m_set


def calc_total_signal(filepath, filename):
    """
    Calculate the total signal intensity of each frame for the whole
    video (tiff stack).

    Parameters
    ----------
    filepath : str
        Path to the tiff file.
    filename : str
        Name of the tiff file.

    Returns
    -------
    total_signal : 1darray
        Signal intensity of each frame ordered by the frame number.
    """
    imstack = tiff.TiffFile(filepath + '/' + filename)
    mvlength = len(imstack.pages)
    total_signal = np.zeros(mvlength)

    for frame_num in range(mvlength):
        im = tiff.imread(filepath + '/' + filename, key=frame_num)
        total_signal[frame_num] = sum(sum(im))
    imstack.close()
    return total_signal


def cut_frames(signal_level, fbc=0.04):
    """
    Find the list of frame number to cut the whole signal plot into seperate
    blocks based on the change of total signal intensity.

    Parameters
    ----------
    signal_level : 1darray
        Signal change over time (can be derived from 'calc_total_signal').
    fbc : float
        The fraction of signal decrease within each block compared to the
        total signal decrease.

    Returns
    -------
    bounds : list of int
        Signal intensities on the boundary of each block.
    frame_lst : list of int
        Frame number where to cut the whole signal plot into blocks.

    Notes
    -----
    The number of blocks is the inverse of the bleaching correction factor,
    fbc. For instance, if fbc=0.04, it means that in each block, the
    decrease in signal intensity is 4% if the total decrease, and the whole
    plot / video is cut into 25 blocks. For some data, it is possible that
    the maximun signal intensity does not appear at the beginning of the
    signal plot. Here, we consider all frames before the maximum in the
    same block as the maximum frame / intensity since usually the number
    of frames is not too large. The user can add in extra blocks if the
    initial intensity is much smaller than the maximum.
    """
    max_intensity, min_intensity = np.max(signal_level), np.min(signal_level)
    frame_num = np.argmax(signal_level)
    total_diff = max_intensity - min_intensity
    block_num = math.ceil(1 / fbc)
    frame_lst = []
    # lower bound of intensity for each block
    bounds = [int(max_intensity - total_diff * i * fbc)
              for i in range(1, block_num + 1)]
    i = 0
    while frame_num < len(signal_level) and i <= block_num:
        if signal_level[frame_num] < bounds[i]:
            frame_lst.append(frame_num)
            frame_num += 1
            i += 1
        else:
            frame_num += 1
    frame_lst = [0] + frame_lst + [len(signal_level)]
    bounds = [int(max_intensity)] + bounds

    return bounds, frame_lst


def min_image(filepath, filename, frames=[]):
    """
    Get the minimum image for a video file (tiff stack), either for the
    whole video or for user defined frames.

    Parameters
    ----------
    filepath : str
        Path to the tiff file.
    filename : str
        Name of the tiff file.
    frames : list of int, optional
        Start and end frame number if not the whole video is used.

    Returns
    -------
    min_im : ndarray
        The minimum image.
    """
    imstack = tiff.TiffFile(filepath + '/' + filename)
    xdim, ydim = np.shape(imstack.pages[0])
    min_im = tiff.imread(filepath + '/' + filename, key=0)

    if not frames:
        frames = [0, len(imstack.pages)]
    for frame_num in range(frames[0], frames[1]):
        im = tiff.imread(filepath + '/' + filename, key=frame_num)
        min_im = (min_im <= im) * min_im + (min_im > im) * im
    imstack.close()
    return min_im


def correct_bleaching(filepath, filename, fbc=0.04, smooth_kernel=251,
                      save_option=True, return_option=False):
    """
    Performs bleaching correction on a tiff image (stack).

    Parameters
    ----------
    fbc : float
        The fraction of signal decrease within each block compared
        to the total signal decrease. Only used when bleach correction
        is True.
    smooth_kernel : int
        The size of the median filter window.
    save_option : bool
        Whether to save the corrected images into tiff files.
    return_option : bool
        Whether to return the corrected image series as a 3d array.

    Returns
    -------
    bc_im : ndarray
        All bleaching-corrected imagee in a 3d array.
    """
    # calculate total signal per frame as a function of time
    sig_b = calc_total_signal(filepath, filename)
    # filter the total signal to get a smoother bleaching profile
    filtered_sig_b = filtering.med_smooth(sig_b, kernel_size=251)
    # calculate the block boundary frame indexes based on the bleaching correction factor fbc
    bounds, frame_lst = cut_frames(filtered_sig_b, fbc=fbc)
    # obtain the total number of blocks.
    block_num = math.ceil(1 / fbc)

    # loop over all the blocks
    for i in range(block_num):
        # calculate the average image for each block
        ave_im = average_image(filepath, filename,
                               frames=[frame_lst[i], frame_lst[i + 1]])
        # loop over every image within the block
        for frame_num in range(frame_lst[i], frame_lst[i + 1]):
            im = tiff.imread(filepath + '/' + filename, key=frame_num)  # read out the image
            norm_im = im - ave_im  # subtract the average from the image
            norm_im = np.int_(np.around(norm_im))  # covnert to integer
            tiff.imwrite(filepath + '/' + filename[:-4] + '_bc0.tif',
                         norm_im, dtype='int', append=True)  # write out a temporariy tiff stack.
            # note that there can be negative values in the image because this is after mean-subtraction

    # store the filename of this temporary tiff stack.
    filename = filename[:-4] + '_bc0.tif'
    # obtain the minimum value across the entire temporary tiff stack (which will be a negative number)
    min_im = min_image(filepath, filename)
    # prepare return value variable
    imstack = tiff.TiffFile(filepath + '/' + filename)
    xdim, ydim = np.shape(imstack.pages[0])
    mvlength = len(imstack.pages)
    if return_option is True:
        bc_im = np.zeros((mvlength, xdim, ydim))
    imstack.close()
    # now need to shift the image pixel values up to set the minimum value to be zero in the following for-loop
    for i in range(mvlength):
        im = tiff.imread(filepath + '/' + filename, key=i)
        norm_im = im - min_im
        norm_im = np.int_(np.around(norm_im))
        if save_option is True:
            tiff.imwrite(filepath + '/' + filename[:-8] + '_bc.tif',
                         norm_im, dtype='int', append=True)
        if return_option is True:
            bc_im[i] = norm_im
    # delete the temperary tiff.
    os.remove(filepath + '/' + filename)
    if return_option is True:
        return bc_im


def moments_all_blocks(filepath, filename,
                       highest_order, smooth_kernel=251, fbc=0.04):
    """
    Get moment-reconstructed images for seperate blocks with user-defined
    noise filtering and bleaching correction factor (fbc).

    Within each block, the amount of signal decrease is identical. This
    amount of signal decrease is characterized by the bleaching correction
    factor, fbc, which is the fractional signal decrease within each block
    (as compared to the total signal decrease over the whole signal plot).

    Parameters
    ----------
    filepath : str
        Path to the tiff file.
    filename : str
        Name of the tiff file.
    highest_order : int
        The highest order number of moment-reconstructed images.
    smooth_kernel : int
        The size of the median filter ('filtering.med_smooth') window.
    fbc : float
        The fraction of signal decrease within each block compared to the
        total signal decrease.

    Returns
    -------
    m_set_all_blocks : dict
        block_number (int) -> {order number (int) -> image (ndarray)}
        A dictionary of moment-reconstructed images of each block.

    Notes
    -----
    Similar to 'calc_moments'. Here we omit previously calculated m_set
    and mean_im as inputs since a block usually has much fewer number of
    frames and takes shorter calculation time.
    """
    all_signal = calc_total_signal(filepath, filename)
    filtered_signal = filtering.med_smooth(all_signal, smooth_kernel)
    _, cut_frame = cut_frames(filtered_signal, fbc)
    block_num = math.ceil(1 / fbc)
    m_set_all_blocks = {}
    for i in range(block_num):
        print('Calculating moments of block %d...' % i)
        m_set_all_blocks[i] = calc_block_moments(filepath,
                                                 filename,
                                                 highest_order,
                                                 [cut_frame[i],
                                                  cut_frame[i + 1]])

    return m_set_all_blocks


def cumulants_all_blocks(m_set_all_blocks):
    """
    Calculate cumulant-reconstructed images from moment-reconstructed images
    of each block. Similar to 'calc_cumulants_from_moments'.
    """
    if m_set_all_blocks == {}:
        raise Exception("'moment_set' is empty.")

    k_set_all_blocks = {i: calc_cumulants_from_moments(m_set_all_blocks[i])
                        for i in m_set_all_blocks}

    return k_set_all_blocks


def block_ave_cumulants(filepath, filename,
                        highest_order, smooth_kernel=251, fbc=0.04):
    """
    Get average cumulant-reconstructed images of all blocks determined by
    user-defined noise filtering and bleaching correction factor (fbc).

    Within each block, the amount of signal decrease is identical. This
    amount of signal decrease is characterized by the bleaching correction
    factor, fbc, which is the fractional signal decrease within each block
    (as compared to the total signal decrease over the whole signal plot).

    Parameters
    ----------
    filepath : str
        Path to the tiff file.
    filename : str
        Name of the tiff file.
    highest_order : int
        The highest order number of moment-reconstructed images.
    smooth_kernel : int
        The size of the median filter ('filtering.med_smooth') window.
    fbc : float
        The fraction of signal decrease within each block compared to the
        total signal decrease.

    Returns
    -------
    ave_k_set : dict
        order number (int) -> image (ndarray)
        A dictionary of avereage cumulant-reconstructed images of all blocks.

    Notes
    -----
    For more information on noise filtering and bleaching corrextion, please
    see appendix 3 of [1].

    References
    ----------
    .. [1] X. Yi, and S. Weiss. "Cusp-artifacts in high order superresolution
    optical fluctuation imaging." bioRxiv: 545574 (2019).
    """
    k_set_all_blocks = cumulants_all_blocks(
        moments_all_blocks(filepath, filename,
                           highest_order,
                           smooth_kernel, fbc))
    block_num = len(k_set_all_blocks)
    k_set_lst = list(k_set_all_blocks.values())
    counter = collections.Counter()
    for d in k_set_lst:
        counter.update(d)
    ave_k_set = dict(counter)
    ave_k_set = {i: ave_k_set[i] / block_num for i in ave_k_set}

    return ave_k_set


def block_ave_moments(filepath, filename,
                      highest_order, smooth_kernel=251, fbc=0.04):
    """
    Get average moment-reconstructed images of all blocks determined by
    user-defined noise filtering and bleaching correction factor (fbc).
    Similar to 'block_ave_cumulants'.

    Parameters
    ----------
    filepath : str
        Path to the tiff file.
    filename : str
        Name of the tiff file.
    highest_order : int
        The highest order number of moment-reconstructed images.
    smooth_kernel : int
        The size of the median filter ('filtering.med_smooth') window.
    fbc : float
        The fraction of signal decrease within each block compared to the
        total signal decrease.

    Returns
    -------
    ave_m_set : dict
        order number (int) -> image (ndarray)
        A dictionary of avereage moment-reconstructed images of all blocks.
    """
    m_set_all_blocks = moments_all_blocks(filepath, filename,
                                          highest_order, smooth_kernel, fbc)
    block_num = len(m_set_all_blocks)
    m_set_lst = list(m_set_all_blocks.values())
    counter = collections.Counter()
    for d in m_set_lst:
        counter.update(d)
    ave_m_set = dict(counter)
    ave_m_set = {i: ave_m_set[i] / block_num for i in ave_m_set}

    return ave_m_set


def calc_xc_im(filepath, filename, ri_range, frames=[]):
    """
    Get cross-cumulant image with different pixel combinations.

    Parameters
    ----------
    filepath : str
        Path to the tiff file.
    filename : str
        Name of the tiff file.
    ri_range : int
        Maximium distance (# pixels) between the original the selected pixel.
    frames : list of int
        The start and end frame number.

    Returns
    -------
    xc : dict
        pixel (tuple) -> image (ndarray)
        A dictionary of reconstructed images with pixel index as keys.
    """
    series_length = ((2 * ri_range + 1) ** 2 - 1) // 2
    ri_series = [(i, j) for i in range(-ri_range, 0)
                 for j in range(-ri_range, ri_range + 1)] + \
                [(0, i) for i in range(-ri_range, 0)]
    imstack = tiff.TiffFile(filepath + '/' + filename)
    xdim, ydim = np.shape(imstack.pages[0])
    im_mean = average_image(filepath, filename, frames)
    imI, imJ = {}, {}
    xc = {ri_series[i]: np.zeros((xdim - 2 * ri_range, ydim - 2 * ri_range))
          for i in range(series_length)}
    if frames:
        mvlength = frames[1] - frames[0]
        for frame_num in range(frames[0], frames[1]):
            im = tiff.imread(filepath + '/' + filename, key=frame_num) - \
                 im_mean
            for series, xc_im in xc.items():
                imI[series] = im[ri_range + series[0]:xdim - ri_range + series[0],
                              ri_range + series[1]:ydim - ri_range + series[1]]
                imJ[series] = im[ri_range - series[0]:xdim - ri_range - series[0],
                              ri_range - series[1]:ydim - ri_range - series[1]]
                xc[series] = xc_im + np.multiply(imI[series], imJ[series])
    else:
        mvlength = len(imstack.pages)
        for frame_num in range(mvlength):
            im = tiff.imread(filepath + '/' + filename, key=frame_num) - \
                 im_mean
            for series, xc_im in xc.items():
                imI[series] = im[ri_range + series[0]:xdim - ri_range + series[0],
                              ri_range + series[1]:ydim - ri_range + series[1]]
                imJ[series] = im[ri_range - series[0]:xdim - ri_range - series[0],
                              ri_range - series[1]:ydim - ri_range - series[1]]
                xc[series] = xc_im + np.multiply(imI[series], imJ[series])
    for series, xc_im in xc.items():
        xc[series] = xc_im / mvlength

    return xc


def calc_moments_with_lag(filepath, filename, tauSeries, frames=[]):
    """
    Get moment-reconnstructed images with defined time lags.

    Parameters
    ----------
    filepath : str
        Path to the tiff file.
    filename : str
        Name of the tiff file.
    tauSeries : list of int
        A list of time lags for frames contribute to moment reconstruction.
        The first element is recommended to be 0, and there should seven
        elements in the list.
    frames : list of int
        The start and end frame number.

    Returns
    -------
    m_set : dict
        order (int) -> dict (partition (tuple) -> image (ndarray))
        A nested dictionary of moment-reconstructions of all partitions.
    """
    if len(tauSeries) < 2:
        raise Exception("The number of time lags should be more than one.")
    if tauSeries[0] != 0:
        raise Exception("The first time lag should be zero.")
    max_order = len(tauSeries)
    mean_im = average_image(filepath, filename, frames)
    xdim, ydim = np.shape(mean_im)

    # define frame range based on tauSeries
    if frames:
        start_frame, end_frame = frames[0], frames[1] - max(tauSeries)
        mvlength = end_frame - start_frame
    else:
        imstack = tiff.TiffFile(filepath + '/' + filename)
        mvlength = len(imstack.pages)
        start_frame, end_frame = 0, mvlength - max(tauSeries)
        mvlength = end_frame - start_frame

    # calculate partitions and their combinations
    seq = list(range(len(tauSeries)))
    if max_order > 3:
        partitions = {element_num: [subset
                                    for subset in itertools.combinations(seq, element_num)]
                      for element_num in range(2, max_order - 1)}

    # set up dictionaries and initialize with zeros array
    m_set = {order: {} for order in range(2, max_order + 1)}
    for order in range(2, max_order + 1):
        m_set[order][tuple(range(0, order))] = np.zeros((xdim, ydim))
    if max_order > 3:
        for order in range(2, max_order - 1):
            m_set[order] = {partitions[order][i]: np.zeros(
                (xdim, ydim)) for i in range(len(partitions[order]))}

    # calculate moments
    for frame_num in range(start_frame, end_frame):
        if max_order > 3:
            # 1. partitions
            for order in range(2, max_order - 1):
                for im_group in partitions[order]:
                    imSeries = np.array([tiff.imread(
                        filepath + '/' + filename,
                        key=frame_num + tauSeries[i]) - mean_im
                                         for i in im_group])
                    m_set[order][im_group] = m_set[order][im_group] + \
                                             np.prod(imSeries, axis=0) / mvlength
            # 2. the whole lag series
            for order in range(max_order - 1, max_order + 1):
                imSeries = np.array([tiff.imread(
                    filepath + '/' + filename, key=frame_num + i) - mean_im
                                     for i in tauSeries])
                m_set[order][tuple(range(0, order))] = m_set[order][tuple(
                    range(0, order))] + \
                                                       np.prod(imSeries[:order], axis=0) / mvlength
        else:
            # only the whole lag series
            imSeries = np.array([tiff.imread(
                filepath + '/' + filename, key=frame_num + i) - mean_im
                                 for i in tauSeries])
            for order in range(2, max_order + 1):
                m_set[order][tuple(range(0, order))] = m_set[order][tuple(
                    range(0, order))] + \
                                                       np.prod(imSeries[:order], axis=0) / mvlength

    return m_set


def calc_cumulants_from_moments_with_lag(m_set, tauSeries):
    """
    Calculate cumulant-reconstructed images from moment-reconstructed images
    with time lags.

    Parameters
    ----------
    m_set : dict
        order (int) -> dict (partition (tuple) -> image (ndarray))
        A nested dictionary of moment-reconstructions of all partitions that
        calculated from 'calc_moments_with_lag'.
    tauSeries : list of int
        A list of time lags for frames contribute to moment reconstruction.
        The first element is recommended to be 0, and there should be seven
        elements in the list. The values should be the same as the input for
        'calc_moments_with_lag'.

    Returns
    -------
    k_set : dict
        order number (int) -> image (ndarray)
        A dictionary of calcualted cumulant-reconstructed images with time
        lags.
    """
    k_set = {}
    seq = list(range(len(tauSeries)))
    max_order = len(tauSeries)
    for order in range(2, max_order + 1):
        k_set[order] = m_set[order][tuple(range(0, order))]
    if max_order > 3:
        for order in range(4, max_order + 1):
            seq = list(range(order))
            max_partition_num = max_order // 2
            for partition_num in range(2, max_partition_num + 1):
                partition_comb = sorted_k_partitions(seq, partition_num)
                for part in partition_comb:
                    k_set[order] = k_set[order] + \
                                   (-1) ** (partition_num - 1) * \
                                   np.math.factorial(partition_num - 1) * \
                                   np.prod([m_set[len(i)][i] for i in part],
                                           axis=0)

    return k_set


from scNodes.core.node import *
from scNodes.core import particlefitting as pfit


def create():
    return ParticleFittingNode()


class ParticleFittingNode(Node):
    title = "PSF fitting"
    group = "PSF-fitting reconstruction"
    colour = (230 / 255, 98 / 255, 13 / 255, 1.0)
    size = 300
    sortid = 1001
    RANGE_OPTIONS = ["All frames", "Current frame only", "Custom range"]
    ESTIMATORS = ["Least squares (GPU)", "Maximum likelihood (GPU)", "No estimator (CPU)"]
    PSFS = ["Gaussian", "Elliptical Gaussian"]

    def __init__(self):
        super().__init__()

        # Set up connectable attributes
        self.dataset_in = ConnectableAttribute(ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.INPUT, parent=self)
        self.localizations_in = ConnectableAttribute(ConnectableAttribute.TYPE_COORDINATES, ConnectableAttribute.INPUT, parent=self)
        self.reconstruction_out = ConnectableAttribute(ConnectableAttribute.TYPE_RECONSTRUCTION, ConnectableAttribute.OUTPUT, parent=self)

        self.range_option = 1
        self.range_min = 0
        self.range_max = 1
        self.estimator = 1
        self.crop_radius = 3
        self.initial_sigma = 1.6
        self.fitting = False
        self.n_to_fit = 1
        self.n_fitted = 0
        self.frames_to_fit = list()
        self.particle_data = ParticleData()

        self.time_start = 0
        self.time_stop = 0

        self.intensity_min = 100.0
        self.intensity_max = -1.0
        self.sigma_min = 1.0
        self.sigma_max = 10.0
        self.offset_min = 0.0
        self.offset_max = -1.0

        self.custom_bounds = False

    def render(self):
        if super().render_start():
            self.dataset_in.render_start()
            self.reconstruction_out.render_start()
            self.dataset_in.render_end()
            self.reconstruction_out.render_end()
            imgui.new_line()
            self.localizations_in.render_start()
            self.localizations_in.render_end()

            imgui.spacing()
            imgui.separator()
            imgui.spacing()
            imgui.set_next_item_width(200)
            _c, self.estimator = imgui.combo("Estimator", self.estimator, ParticleFittingNode.ESTIMATORS)
            if self.estimator in [0, 1]:
                imgui.same_line()
                imgui.button("?", 19, 19)
                self.tooltip("This node uses pyGpufit, a GPU fitting library by Przybylski et al. - original publication:\n"
                             "Przybylski et al. (2017) Gpufit: An open-source toolkit for GPU-accelerated curve fitting.\n"
                             "Sci. Rep. 7:15722. doi: 10.1038/s41598-017-15313-9")
            self.any_change = _c or self.any_change
            imgui.push_item_width(80)
            if self.estimator in [0, 1]:
                _c, self.initial_sigma = imgui.input_float("Initial sigma (px)", self.initial_sigma, 0, 0, "%.1f")
                self.any_change = _c or self.any_change
                _c, self.crop_radius = imgui.input_int("Fitting radius (px)", self.crop_radius, 0, 0)
                self.any_change = _c or self.any_change
            imgui.pop_item_width()
            imgui.spacing()
            imgui.set_next_item_width(200)
            _c, self.range_option = imgui.combo("Range", self.range_option,
                                                ParticleFittingNode.RANGE_OPTIONS)
            if self.range_option == 2:
                imgui.push_item_width(80)
                _c, (self.range_min, self.range_max) = imgui.input_int2('[start, stop) index', self.range_min,
                                                                        self.range_max)
                imgui.pop_item_width()
            imgui.spacing()
            if self.fitting:
                self.progress_bar(self.n_fitted / self.n_to_fit)
                imgui.spacing()
                imgui.spacing()
                imgui.spacing()
                imgui.spacing()

            _play_btn_clicked, _state = self.play_button()
            if _play_btn_clicked and _state:
                self.init_fit()
            elif _play_btn_clicked and self.fitting:
                self.fitting = False
                self.particle_data.bake()

            ## TODO: add camera offset and ADU per photon (to LOAD DATASET?)
            if not self.particle_data.empty:
                imgui.text("Reconstruction info")
                imgui.text(f"particles: {len(self.particle_data.particles)}")
                if not self.fitting:
                    imgui.text(f"x range: {self.particle_data.x_min / 1000.0:.1f} to {self.particle_data.x_max / 1000.0:.1f} um")
                    imgui.text(f"y range: {self.particle_data.y_min / 1000.0:.1f} to {self.particle_data.y_max / 1000.0:.1f} um")
            imgui.spacing()
            header_expanded, _ = imgui.collapsing_header("Advanced", None)
            if header_expanded:
                self.render_advanced()

            super().render_end()

    def render_advanced(self):
        _c, self.custom_bounds = imgui.checkbox("Use custom parameter bounds", self.custom_bounds)
        Node.tooltip("Edit the bounds for particle parameters intensity, sigma, and offset.")
        if self.custom_bounds:
            imgui.push_item_width(45)
            _c, self.sigma_min = imgui.input_float("min##sigma", self.sigma_min, 0, 0, "%.1f")
            Node.tooltip("Enter value to limit particle sigma to a minimum value, or '-1.0' to leave unbounded.")
            imgui.same_line()
            _c, self.sigma_max = imgui.input_float("max##sigma", self.sigma_max, 0, 0, "%.1f")
            Node.tooltip("Enter value to limit particle sigma to a maximum value, or '-1.0' to leave unbounded.")
            imgui.same_line()
            imgui.text("sigma (px)")

            _c, self.intensity_min = imgui.input_float("min##int", self.intensity_min, 0, 0, "%.1f")
            Node.tooltip("Enter value to limit particle intensity to a minimum value, or '-1.0' to leave unbounded.")
            imgui.same_line()
            _c, self.intensity_max = imgui.input_float("max##int", self.intensity_max, 0, 0, "%.1f")
            Node.tooltip("Enter value to limit particle intensity to a maximum value, or '-1.0' to leave unbounded.")
            imgui.same_line()
            imgui.text("intensity (counts)")

            _c, self.offset_min = imgui.input_float("min##offset", self.offset_min, 0, 0, "%.1f")
            Node.tooltip("Enter value to limit particle offset to a minimum value, or '-1.0' to leave unbounded.")
            imgui.same_line()
            _c, self.offset_max = imgui.input_float("max##offset", self.offset_max, 0, 0, "%.1f")
            Node.tooltip("Enter value to limit particle offset to a maximum value, or '-1.0' to leave unbounded.")
            imgui.same_line()
            imgui.text("offset (counts)")

            imgui.pop_item_width()

    def init_fit(self):
        try:
            self.time_start = time.time()
            dataset_source = Node.get_source_load_data_node(self)
            self.particle_data = ParticleData(dataset_source.pixel_size)
            self.particle_data.set_reconstruction_roi(np.asarray(dataset_source.dataset.reconstruction_roi) * dataset_source.pixel_size)
            self.fitting = True
            self.frames_to_fit = list()
            if self.range_option == 0:
                dataset = dataset_source.dataset
                n_frames = dataset.n_frames
                self.frames_to_fit = list(range(0, n_frames))
            elif self.range_option == 2:
                self.frames_to_fit = list(range(self.range_min, self.range_max))
            elif self.range_option == 1:
                dataset = dataset_source.dataset
                self.frames_to_fit = [dataset.current_frame]
            self.n_to_fit = len(self.frames_to_fit)
            self.n_fitted = 0
        except Exception as e:
            self.fitting = False
            cfg.set_error(e, "Error in init_fit: "+str(e))

    def on_update(self):
        try:
            if self.fitting:
                if len(self.frames_to_fit) == 0:
                    self.fitting = False
                    self.play = False
                    self.particle_data.bake()
                    print(self.particle_data)
                else:
                    fitted_frame = self.get_image(self.frames_to_fit[-1])
                    self.n_fitted += 1
                    if fitted_frame is not None:
                        particles = fitted_frame.particles
                        self.frames_to_fit.pop()
                        if not fitted_frame.discard:
                            self.particle_data += particles
        except Exception as e:
            self.fitting = False
            self.play = False
            cfg.set_error(e, "Error in ParticleFitNode.on_update(self): "+str(e))

    def get_image_impl(self, idx=None):
        data_source = self.dataset_in.get_incoming_node()
        coord_source = self.localizations_in.get_incoming_node()
        if data_source and coord_source:
            frame = data_source.get_image(idx)
            coordinates = coord_source.get_coordinates(idx)
            frame.maxima = coordinates
            if frame.discard:
                frame.particles = None
                return frame
            particles = list()
            if self.estimator in [0, 1]:
                particles = pfit.frame_to_particles(frame, self.initial_sigma, self.estimator, self.crop_radius, constraints = [self.intensity_min, self.intensity_max, -1, -1, -1, -1, self.sigma_min, self.sigma_max, self.offset_min, self.offset_max])
            elif self.estimator == 2:
                particles = list()
                pxd = frame.load()
                for i in range(len(frame.maxima)):
                    intensity = pxd[frame.maxima[i, 0], frame.maxima[i, 1]]
                    x = frame.maxima[i, 1]
                    y = frame.maxima[i, 0]
                    particles.append(Particle(idx, x, y, 1.0, intensity))
            frame.particles = particles
            new_maxima = list()
            for particle in frame.particles:
                new_maxima.append([particle.y, particle.x])
            frame.maxima = new_maxima
            return frame

    def get_particle_data_impl(self):
        self.particle_data.clean()
        return self.particle_data

    def pre_save_impl(self):
        pass # cfg.pickle_temp["particle_data"] = self.particle_data
        # self.particle_data = ParticleData()

    def post_save_impl(self):
        pass # self.particle_data = cfg.pickle_temp["particle_data"]





import subprocess
import sys
import os
import tkinter as tk
from tkinter import filedialog
tkroot = tk.Tk()
tkroot.withdraw()

def install():
    root = os.path.join(os.path.dirname(__file__)).replace("\\", "/") + "/"
    #directory = filedialog.askdirectory() + "/"
    #executable = sys.executable
    #with open(directory + "scNodes.bat", "w") as f:
    #    f.write('@echo off\n"' + str(executable) + '" "' + root + '__main__.py' + '"\npause')
    try:
        print("Trying to install pyGpufit, with wheel at:\n"+root+"wheels/pyGpufit-1.2.0-py2.py3-none-any.whl")
        subprocess.check_call([sys.executable, "-m", "pip", "install", root+"wheels/pyGpufit-1.2.0-py2.py3-none-any.whl", "pyGpufit"])
    except Exception as e:
        print("Installation of pyGpufit failed. Try installing it manually.")
        raise e

if __name__=="__main__":
    install()



from scNodes.core.node import *


def create():
    return FrameShiftNode()


class FrameShiftNode(Node):
    title = "Frame shift"
    group = "Image processing"
    colour = (235 / 255, 232 / 255, 80 / 255, 1.0)
    sortid = 104

    def __init__(self):
        super().__init__()
        self.size = 140

        # Set up connectable attributes
        self.dataset_in = ConnectableAttribute(ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.INPUT, parent = self)
        self.dataset_out = ConnectableAttribute(ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.OUTPUT, parent = self)
        self.shift = 0

    def render(self):
        if super().render_start():
            self.dataset_out.render_start()
            self.dataset_in.render_start()
            self.dataset_out.render_end()
            self.dataset_in.render_end()
            imgui.spacing()
            imgui.separator()
            imgui.spacing()

            imgui.push_item_width(80)
            _c, self.shift = imgui.input_int("shift", self.shift, 1, 10)
            self.any_change = _c or self.any_change

            super().render_end()

    def get_image_impl(self, idx=None):
        data_source = self.dataset_in.get_incoming_node()
        if data_source:
            return data_source.get_image(idx + self.shift)



from scNodes.core.node import *
from scNodes.nodes.pysofi import pysofi
from scNodes.nodes.pysofi import finterp
import os
import tifffile

def create():
    return SofiNode()


class SofiNode(Node):
    title = "SOFI reconstruction"
    group = ["Alternative reconstruction"]
    colour = (98 / 255, 13 / 255, 230 / 255, 1.0)  # (230 / 255, 98 / 255, 13 / 255, 1.0)
    size = 245
    sortid = 1006

    RANGE_OPTIONS = ["Full stack", "Selected range"]
    OUTPUT_OPTIONS = ["Moments", "Cumulants"]

    def __init__(self):
        super().__init__()

        # defining in- and output attributes.
        self.dataset_in = ConnectableAttribute(ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.INPUT, parent=self)
        self.image_out = ConnectableAttribute(ConnectableAttribute.TYPE_IMAGE, ConnectableAttribute.OUTPUT, parent=self)

        self.result = None
        self.range_option = 1
        self.range_min = 0
        self.range_max = 100
        self.output_option = 1
        self.moment_order = 4
        self.cumulant_order = 4
        self.magnification = 2

        self.lut = "Heatmap"
        self.pysd = None
        self.pysd_current = False
        self.cumulant_imgs = dict()
        self.cumulants_up_to_date = False
        self.prev_cumulant_settings = ""
        self.sofi_img = None
        self.stack = None
        self.frames_requested = list()
        self.n_frames_processed = 0
        self.n_frames_requested = 1
        self.pxsizein = 100
        self.stack_prepared = False
        self.processing = False
        self.temp_dir = "_srnodes_temp_dir_sofinode_"+str(self.id)

    def render(self):
        if super().render_start():  # as in the above __init__ function, the render function must by calling the base class' render_start() function, and end with a matching render_end() - see below.
            self.dataset_in.render_start()  # calling a ConnectableAttribute's render_start() and render_end() handles the rendering and connection logic for that attribute.
            self.image_out.render_start()
            self.dataset_in.render_end()
            self.image_out.render_end()  # callin start first for both attributes and end afterwards makes the attributes appear on the same line / at the same height.

            imgui.spacing()  # add a small vertical whitespace
            imgui.separator()  # draw a line separating the above connectors from the rest of the body of the node. purely visual.
            imgui.spacing()



            imgui.set_next_item_width(120)
            _c, self.output_option = imgui.combo("Output type", self.output_option, SofiNode.OUTPUT_OPTIONS)
            imgui.same_line()
            imgui.button("?", width=20, height=20)
            self.tooltip("This node is based on PySOFI, see:\n"
                         "Miao et al. 'PySOFI: an open source Python package for SOFI.' (2022)\n\n"
                         ""
                         "During PySOFI processing, GUI will be unresponsive - processing time\n"
                         "varies depending on size of the dataset. Crop images or reduce range\n"
                         "of frames to reduce processing time.")
            imgui.push_item_width(120)
            self.any_change = self.any_change or _c
            if self.output_option == 0:
                _c, self.moment_order = imgui.input_int("Moment order", self.moment_order, 1, 1)
                self.any_change = self.any_change or _c
            elif self.output_option == 1:
                _c, self.cumulant_order = imgui.input_int("Cumulant order", self.cumulant_order, 1, 1)
                self.any_change = self.any_change or _c

            imgui.set_next_item_width(120)
            _c, self.range_option = imgui.combo("Range", self.range_option, SofiNode.RANGE_OPTIONS)
            self.any_change = self.any_change or _c

            if self.range_option == 1:
                imgui.text("Range:")
                imgui.push_item_width(30)
                _c, self.range_min = imgui.input_int("start", self.range_min, 0, 0)
                self.any_change = self.any_change or _c
                imgui.same_line()
                _c, self.range_max = imgui.input_int("stop", self.range_max, 0, 0)
                self.any_change = self.any_change or _c
                imgui.pop_item_width()

            imgui.pop_item_width()
            imgui.set_next_item_width(60)
            _c, self.magnification = imgui.input_int("Magnification", self.magnification, 0, 0)
            self.magnification = max([1, self.magnification])
            self.any_change = _c or self.any_change

            if self.processing:
                imgui.text("Preparing stack")
                self.progress_bar(self.n_frames_processed / self.n_frames_requested)
                imgui.spacing()
                imgui.spacing()
                imgui.spacing()
                imgui.spacing()

            _cw = imgui.get_content_region_available_width()
            imgui.new_line()
            imgui.same_line(position=(_cw - 100) / 2)
            if imgui.button("Reconstruct", width=100, height=30):
                self.init_reconstruction()
            super().render_end()

    def init_reconstruction(self):
        # unfortunately pysofi works on files saved on disk, not in RAM. numpy arrays as input would be more efficient, but
        # that would require editing the pysofi source which want to avoid. So the first step in building a sofi reconstruction
        # is to get the required data as a numpy array, and save it to disk. This is a bit inefficient.

        # If the current stack range settings are the same as before, skip grabbing the stack.
        self.pysd_current = False

        if not os.path.isdir(self.temp_dir):
            os.mkdir(self.temp_dir)
        if self.range_option == 0:
            dataset_source = Node.get_source_load_data_node(self)
            self.frames_requested = list(range(0, dataset_source.dataset.n_frames))
            self.range_min = 0
        else:
            self.frames_requested = list(range(self.range_min, self.range_max))
        sample_input_img = self.dataset_in.get_incoming_node().get_image_impl(idx=0)
        width, height = sample_input_img.load().shape
        self.pxsizein = sample_input_img.pixel_size
        self.stack = np.zeros((len(self.frames_requested), width, height))
        self.frames_requested.reverse()
        self.stack_prepared = False
        self.processing = True
        self.n_frames_requested = len(self.frames_requested)
        self.n_frames_processed = 0

    def on_update(self):
        if self.processing and not self.stack_prepared:
            datasource = self.dataset_in.get_incoming_node()
            self.stack[self.frames_requested[0] - self.range_min, :, :] = datasource.get_image(idx = self.frames_requested[0]).load()
            self.frames_requested.pop()
            self.n_frames_processed += 1
            if not self.frames_requested:
                self.stack_prepared = True
                # save stack:
                with tifffile.TiffWriter(self.temp_dir + "/sofistack.tif") as stack_out:
                    for i in range(self.stack.shape[0]):
                        stack_out.write(self.stack[i])
                # then prep the reconstruction
                self.pysd = pysofi.PysofiData(self.temp_dir, "sofistack.tif")
                self.pysd_current = True
                self.build_reconstruction()

    def build_reconstruction(self):
        if self.output_option == 0:
            self.sofi_img = self.pysd.moment_image(order=self.moment_order, finterp=self.magnification > 1, interp_num=self.magnification)
        elif self.output_option == 1:
            self.cumulant_imgs = self.pysd.cumulants_images(highest_order=self.cumulant_order)
            if self.magnification > 1:
                self.sofi_img = finterp.fourier_interp_array(self.cumulant_imgs[self.cumulant_order], [self.magnification])[0]
            else:
                self.sofi_img = self.cumulant_imgs[self.cumulant_order]
        self.any_change = True
        self.processing = False

    def get_image_impl(self, idx):
        if self.sofi_img is not None:
            outframe = Frame("sofi node output image")
            outframe.data = self.sofi_img
            outframe.pixel_size = self.pxsizein / self.magnification
            return outframe
        else:
            return None
